import os
try:
    from pathlib import Path
except:    
    from pathlib2 import Path
import argparse
import sys
import itertools
import geopandas as gpd
from datetime import datetime, timedelta
import copy
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.dates import date2num
import re
from matplotlib.ticker import ScalarFormatter, FormatStrFormatter
import numpy as np
import pandas as pd
from scipy import optimize
import fire
import warnings
try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        from tqdm import tqdm_notebook as tqdm # Jupyter notebook or qtconsole        
    elif shell == 'TerminalInteractiveShell':
        from tqdm import tqdm # Terminal running IPython
    else:
        from tqdm import tqdm # Other type (?)
except NameError:
    from tqdm import tqdm # Probably standard Python interpreter
    
#for debugging within notebook    

try:
    from IPython.core.debugger import set_trace    
except:
    pass

# for python 2/3 compatibility
try:
    xrange
except NameError:
    xrange = range    
try:
    get_ipython().magic('matplotlib inline')    
except:
    pass
from hkvsobekpy.io.his import __his_class
from hkvsobekpy.io.bui import __bui_class
from hkvsobekpy.core.utils import *
his_reader = __his_class()
bui_reader = __bui_class()

class __plausibility_class(object):

    def __init__(self):
        """
        SobekResultaten class. Class met methodes om een .his bestand 
        uit te lezen.
        WaterstandStatistiek class. Class met methodes om een GumbleFit te bepalen
        """
        # initatie van lege immutable objects
        # zie: https://stackoverflow.com/a/22526544/2459096
        # python2.7: obj = argparse.Namespace()
        # python3.x: obj = types.SimpleNamespace()        
        # self.stats = argparse.Namespace()
        # self.stats.plotposities = argparse.Namespace()
        # self.stats.stap1 = argparse.Namespace()
        # self.stats.stap2 = argparse.Namespace()
        # self.stats.stap3 = argparse.Namespace()
        # self.stats.stap4 = argparse.Namespace()
        
    class __errors__(object):
        """
        error class met verschillende foutmeldingen
        """
        @staticmethod
        def gewogenGemiddeldeError():
            raise ValueError('Kan gewogen gemiddelde niet bepalen. Vereiste is minimaal 2 punten aan de linker en rechterzijde van de gekozen T')
        def gebeurtenissenError():
            raise ValueError('Meer gebeurtenissen dan N. Zorg ervoor dat je tijdreeks slechts 1 gebeurtenis per jaar heeft')
        def schrijfTabelError():
            raise ValueError('Kan deze tabel alleen wegschrijven met T10,25,50,100 voor Gumbel en T10 voor gewogen gemiddelde')
        
    def prepare_bui_his(self, df_his, df_bui, bui_locations='auto', bui_locations_aantal=10):
        """
        Prepare parameters from bui- and his-file for plotting purposes
        
        Parameters
        ----------
        his_file : str
            name of his-file to query (e.g: 'reachseg', 'calcpnt', 'struc')
        location : str
            location to query in his-file
        parameter : str
            parameter to query in his-file
        his_folder : str
            path to folder containing the his_file
        df_bui :         
        bui_locations : str or list
            welke bui_locations worden gebruikt om een selectie te maken in de
            bui-file, if 'bui_locations' is 'auto', dan worden de eerste x locations gebruikt
            (waar x = 'bui_locations_aantal')    
        bui_locations_aantal : int
            aantal locations mee te nemen tijdens plotten wanneer 'auto' is 
            opgegeven voor parameter 'bui_locations' (default: 10)
        
        Returns
        -------
        df_his : pandas.DataFrame
            timeseries containing all available timesteps in his-file
        df_bui_sel : pandas.DataFrame
            timeseries containing selection of bui-file, aligning with his-file
        start_his : datetime
            first datetime step of his-file
        end_his : datetime
            last datetime step of his-file
        df_bui_std : float
            standard deviation of timeseries selection bui-file
        df_his_std : float
            standard deviation of timeseries his-hile
        """
      
        start_his  = df_his.index[0]
        end_his    = df_his.index[-1]    
        df_bui_sel = df_bui.loc[start_his-pd.DateOffset(1):end_his+pd.DateOffset(1)]
        if df_bui_sel.values.size == 0:
            raise ValueError("""
            cannot slice bui-file, seems time-period does not align with his-file
            time-period his-file: {0} - {1}
            time-period bui-file: {2} - {3}    
            """.format(df_his.index[0], df_his.index[-1], df_bui.index[0], df_bui.index[-1]))  
            
    #     # create dummy waarden voor df_bui_sel
    #     for loc in df_bui_sel.columns:
    #         df_bui_sel.loc[df_bui_sel[loc] == 0,loc] = df_bui_sel[loc].apply(lambda x: np.random.normal(0,1))   
         
        if bui_locations == 'auto':
            df_bui_sel = df_bui_sel.iloc[:, 0:bui_locations_aantal]
        if isinstance(bui_locations, list):
            df_bui_sel = df_bui_sel.loc[:, bui_locations]

        df_his_std = np.std(df_his.values)
        df_bui_std = np.std(df_bui_sel.values)
        if df_bui_std == 0.:
            df_bui_std = 0.5
        if df_his_std == 0.:
            df_his_std = 0.5 
        
        return df_bui_sel, start_his, end_his, df_bui_std, df_his_std
        
    def plot_bui_his(self, df_his, df_bui_sel, start_his, end_his, df_bui_std, df_his_std, his_file, parameter, location, out_folder, savefigure, barthreshold=4, barwidth = 0.02, baroffset=0.01):
        if barthreshold > 3:
            barthreshold = 3
        fig, ax1 = plt.subplots(figsize=(8,6))
        ax2 = ax1.twinx()

        # AXIS 1 :: his-file
        ax1.plot(df_his.index, df_his.values,color='#BFD600',marker='o',lw=4,label=parameter)
        ax1.set_xlim(start_his,end_his)
        ax1.set_ylabel(parameter)
        ax1.tick_params(labelbottom='on',labeltop='off', labelright="off",labelleft='on')
        ax1.set_ylabel(parameter,rotation=90)
        ax1.set_yticks(np.linspace(ax1.get_ybound()[0]-(df_his_std), ax1.get_ybound()[1]+df_his_std, 5))
        ax1.grid(True, axis='y')
        leg=ax1.legend(bbox_to_anchor=(0, 0.02),loc=3,prop={'size':10},ncol=5)
        leg.draw_frame(False)

        if len(df_bui_sel.columns) > barthreshold:
            # AXIS 2 :: bui-file, width=barwidth, Draw line-chart
            colors = [ cm.Blues(x) for x in np.linspace(0.2, 1.0, len(df_bui_sel.columns)) ]
            for y_arr, label, color in zip(df_bui_sel.values.T, df_bui_sel.columns, colors):
                ax2.plot(df_bui_sel.index, y_arr, label='P.{0} (mm)'.format(label), lw=1.5,color=color)
                ax2.fill_between(df_bui_sel.index, 0, y_arr, color=color,alpha=0.05)
            ax2_ylim = ax2.get_ylim()[::-1]
            bottom = max(max(0,ax2_ylim[0]),max(df_bui_std,ax2_ylim[1]))
            top = min(max(0,ax2_ylim[0]),max(df_bui_std,ax2_ylim[1]))
            ax2.set_ylim(top=top, bottom=bottom)
            ax2.set_xlim(start_his,end_his)
            ax2.tick_params(labelbottom='off',labeltop='on', labelright="on",labelleft='off')
            ax2.set_ylabel('Neerslag (mm)',rotation=90)
            ax2.set_yticks(np.linspace(0, ax2.get_ybound()[1]+df_bui_std, 5))
            leg=ax2.legend(bbox_to_anchor=(1.52, 0.02),loc=4,prop={'size':10},ncol=1)
            leg.draw_frame(False)
        
        else:
            print('create a barchart for the bui-locations')
            x = date2num(df_bui_sel.index.to_pydatetime())
            if len(df_bui_sel.columns) == 3:
                colors = [ cm.Blues(x) for x in np.linspace(0.25, 0.8, len(df_bui_sel.columns)) ]
                series0 = df_bui_sel.iloc[:,0].tolist()
                label0 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,0].name)
                series1 = df_bui_sel.iloc[:,1].tolist()
                label1 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,1].name)
                series2 = df_bui_sel.iloc[:,2].tolist()
                label2 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,2].name)

                w = barwidth
                offset = baroffset
                ax2.bar(x - offset, series0, width=w, align='center', label=label0, color=colors[0], alpha=0.70)
                ax2.bar(x, series1, width=w, align='center', label=label1 , color=colors[1], alpha=0.70)
                ax2.bar(x + offset, series2, width=w, align='center', label=label2 , color=colors[2], alpha=0.70)

            if len(df_bui_sel.columns) == 2:
                colors = [ cm.Blues(x) for x in np.linspace(0.35, 0.75, len(df_bui_sel.columns)) ]
                series0 = df_bui_sel.iloc[:,0].tolist()
                label0 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,0].name)
                series1 = df_bui_sel.iloc[:,1].tolist()
                label1 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,1].name)

                w = barwidth
                offset = baroffset
                ax2.bar(x - offset, series0, width=w, align='center', label=label0, color=colors[0], alpha=0.75)
                ax2.bar(x, series1, width=w, align='center', label=label1 , color=colors[1], alpha=0.75)   

            if len(df_bui_sel.columns) == 1:
                colors = [ cm.Blues(x) for x in np.linspace(0.45, 0.55, len(df_bui_sel.columns)) ]
                series0 = df_bui_sel.iloc[:,0].tolist()
                label0 = 'P.{0} (mm)'.format(df_bui_sel.iloc[:,0].name)

                w = barwidth
                offset = baroffset
                ax2.bar(x, series0, width=w, align='center', label=label0 , color=colors[0])                   
               
            ax2.xaxis_date()
            ax2.autoscale(tight=True )   

            ax2_ylim = ax2.get_ylim()[::-1]
            bottom = max(max(0,ax2_ylim[0]),max(df_bui_std,ax2_ylim[1]))
            top = min(max(0,ax2_ylim[0]),max(df_bui_std,ax2_ylim[1]))
            ax2.set_ylim(top=top, bottom=bottom)
            ax2.set_xlim(start_his,end_his)
            ax2.tick_params(labelbottom='off',labeltop='on', labelright="on",labelleft='off')
            ax2.set_ylabel('Neerslag (mm)',rotation=90)
            ax2.set_yticks(np.linspace(0, ax2.get_ybound()[1]+df_bui_std, 5))
            leg=ax2.legend(bbox_to_anchor=(0.95, 0.02),loc=4,prop={'size':10},ncol=1)
            leg.draw_frame(False)            

        # fix zorder
        ax1.set_zorder(ax2.get_zorder()+1) # put ax1 in front of ax2 
        ax1.patch.set_visible(False) # hide the 'canvas'
        #ax2.patch.set_visible(False) # hide the 'canvas'

        plt.title('location: '+location)
        plt.tight_layout()
        if savefigure == False:            
            plt.show()
            return
        else:   
            his_file = "".join(re.findall("[A-Za-z0-9]", his_file))
            location = "".join(re.findall("[A-Za-z0-9]", location))
            parameter = "".join(re.findall("[A-Za-z0-9]", parameter))
            path = os.path.join(out_folder,'{0}_{1}_{2}'.format(his_file,location,parameter))
            #print(path)
            plt.savefig(path, dpi=100, bbox_inches='tight')
            plt.close(fig) 
            return
        
    def table_bui_his(self, df_his, df_bui_sel, savetable=False, his_file='', his_parameter='', his_location='', out_folder=''):
        """
        write table from combined his-file and bui-file
        
        Parameters
        ----------
        df_his : pandas.DataFrame
        df_bui_sel : pandas.DataFrame
        his_file : str
        his_parameter : str
        his_location : str
        out_folder : str
        
        Returns
        -------
        df_bui_his : pandas.DataFrame
        """
        # include new index level on column
        df_bui_sel = pd.concat([df_bui_sel], keys=['Precipitation (mm)'], names=['parameter'], axis=1)
        # add column description
        df_bui_sel.columns.levels[1].name='location'
        # reorder levels of bui-file
        df_bui_sel = df_bui_sel.reorder_levels(['location','parameter'], axis=1)
        # concatenate both dataframes
        df_bui_his = pd.concat((df_his,df_bui_sel), axis=1, join='outer')
        # save to csv
        if savetable==True:
            his_file = "".join(re.findall("[A-Za-z0-9]", his_file))
            his_location = "".join(re.findall("[A-Za-z0-9]", his_location))
            his_parameter = "".join(re.findall("[A-Za-z0-9]", his_parameter))
            path = os.path.join(out_folder,'{0}_{1}_{2}.csv'.format(his_file,his_location,his_parameter))
            #print(path)
            df_bui_his.to_csv(path, encoding='windows-1252')
            return
        else:
            return df_bui_his

        
    def EnsembleRunner(self, shp_file, bui_file, his_folder, out_folder, 
                       shp_hiskey, shp_locationkey, shp_parameterkey, savefigure=True,
                       bui_locations='auto', bui_locations_aantal=10, savetable=True,
                       normalize_by_unicode=True, include_simularity=True, sequence_simularity=0.82,
                       threshold_bar=3, barwidth = 0.02, baroffset=0.01):
        
        """
        Combine timeseries of his files and bui locations using a dbf/shp file 
        containing information about the mapping. Prepares figures and tables in batch function.
        
        Parameters
        ----------
        shp_file : str
            path to dbf/shp file containing three columns to query his-file and location and parameter
        bui_file : str
            path to .bui file, containing a single precipitation event for 1 or more locations
        his_folder : str
            path to folder containing the his_file
        out_folder : str
            path to folder where to store output deriviates
        shp_hiskey : str
            his column key in shp/dbf file (contains name of his-file to query 
            (e.g: 'reachseg', 'calcpnt', 'struc'))
        shp_locationkey : str
            location column key in shp/dbf file
        shp_parameterkey : str
            parameter column key in shp/dbf file
        savefigure : boolean
            options to exlude or include export of figures
        bui_locations : str or list
            welke bui_locations worden gebruikt om een selectie te maken in de
            bui-file, if 'bui_locations' is 'auto', dan worden de eerste x locations gebruikt
            (waar x = 'bui_locations_aantal')    
        bui_locations_aantal : int
            aantal locations mee te nemen tijdens plotten wanneer 'auto' is 
            opgegeven voor parameter 'bui_locations' (default: 10)
        savetable : boolean
            options to exlude or include export of table (csv format) 
        normalize_by_unicode : boolean (default: True)
            inlcude this option to include NFKD unicode compatibility decomposition.
            see: http://unicode.org/reports/tr15/
        include_simluratiy : boolean (default: True)
            include this option to include Ratcliff/Obershelp pattern recognition
        sequence_simularity : float (default 0.82)
            number between 0.0 and 1.0, function as threshold, where only a simularity above 
            this value is mapped   

        Returns
        -------
        None. Figures and tables are prepared in the out_folder
        """
        
       
        print('start ensemble runner')
        gdf = gpd.read_file(shp_file)
        
        print('read shp-file')
        df_bui = bui_reader.read_bui(bui_file)
        print('read bui-file')
        progress_bar = tqdm(gdf.T.columns.tolist())
        for idx in progress_bar:        
            row = gdf.iloc[idx,:]
            progress_bar.set_description("location %s" % row[shp_locationkey].ljust(20))
            his_file = row[shp_hiskey]
            location = row[shp_locationkey]
            parameter = row[shp_parameterkey]            
            
            # read his-file
            df_his, parameter = his_reader.read_series(his_file, location, parameter, his_folder,
                            normalize_by_unicode, include_simularity, sequence_simularity, return_matching_parameter=True)
            
            # prepare files for plotting
            df_bui_sel, start_his, end_his, df_bui_std, df_his_std = self.prepare_bui_his(
                df_his, df_bui, 
                bui_locations=bui_locations, 
                bui_locations_aantal=bui_locations_aantal)
            
            # plot the figure
            self.plot_bui_his(df_his, df_bui_sel, start_his, end_his, 
                            df_bui_std, df_his_std, his_file, parameter, 
                            location, out_folder, savefigure)
                            
            # save the data tables
            self.table_bui_his(df_his, df_bui_sel, savetable=savetable, his_file=his_file, 
                his_parameter=parameter, his_location=location, out_folder=out_folder)
            
        return (print('done'))



import fire
from hkvsobekpy.io.bui import __bui_class
from hkvsobekpy.io.his import __his_class
from hkvsobekpy.core.waterlevelstat import __waterlevelstat_class
from hkvsobekpy.core.plausibility import __plausibility_class

__doc__ = """package for water-statistics and plausibility checker using his- and bui-file"""
__version__ = '1.4'

# initiate class for .bui-files
__bui = __bui_class()
read_bui = __bui.read_bui
write_bui = __bui.write_bui

# initiate class for .his-files
read_his = __his_class()

# initiate class for waterlevelstats
waterlevelstat = __waterlevelstat_class()

# initiate class for plausibilitychecker
plausibility = __plausibility_class()

if __name__ == '__main__':
    fire.Fire()


from difflib import SequenceMatcher
import unicodedata

def similar(a, b):
    return SequenceMatcher(None, a, b).ratio()

def normalize_caseless(text):
    return unicodedata.normalize("NFKD", text.casefold())

def caseless_equal(left, right):
    return normalize_caseless(left) == normalize_caseless(right)
    
def compare_df_column_his_list(df, df_column_key, his_parameters, normalize_by_unicode=True, 
                                include_simularity=False, sequence_simularity=0.82):
    """
    function to apply unicode normalization and similarity checking for two list of columns
    
    Parameters
    ----------
    df : geopandas.GeoDataFrame
        dataframe containing the input columns, normally is the source a shp or dbf file
    df_column_key : str
        name of column for usage as key
    his_parameters : list
        list of strings of locations/parameters of his-file to compare against the column 
        in the geodataframe
    normalize_by_unicode : boolean
        inlcude this option to include NFKD unicode compatibility decomposition.
        see: http://unicode.org/reports/tr15/
    include_simluratiy : boolean
        include this option to include Ratcliff/Obershelp pattern recognition
    sequence_simularity : float
        number between 0.0 and 1.0, function as threshold, where only a simularity above 
        this value is mapped
    
    Returns
    df : geopandas.GeoDataFrame
        dataframe where column matching the column key is updated with matching values 
    
    """

    for df_idx, df_parameter in enumerate(df[df_column_key]):

        for his_parameter in his_parameters:
            # option to have only a unicode normalization
            if normalize_by_unicode == True and include_simularity == False:
                if caseless_equal(his_parameter, df_parameter) == True:
                    df.loc[df_idx, df_column_key] = his_parameter
                    print('{0} changed into {1}'.format(
                        df_parameter, his_parameter))

            # option to apply only parameter similarity functionality
            if include_simularity == True and normalize_by_unicode == False:
                smty = similar(his_parameter, df_parameter)
                # print(smty)
                if sequence_simularity <= smty < 1.0:
                    df.loc[df_idx, df_column_key] = his_parameter
                    print('{0} changed into {1}'.format(
                        df_parameter, his_parameter))

            # option to apply both unicode normalization and similarity functionality
            if include_simularity == True and normalize_by_unicode == True:
                # first check unicode normalization
                if caseless_equal(his_parameter, df_parameter) == True:
                    df.loc[df_idx, df_column_key] = his_parameter
                    print('{0} changed into {1}'.format(
                        df_parameter, his_parameter))
                # if not succeed try similarity
                else:
                    
                    smty = similar(normalize_caseless(his_parameter), normalize_caseless(df_parameter))
                    #print(smty)
                    if sequence_simularity <= smty < 1.0:
                        df.loc[df_idx, df_column_key] = his_parameter
                        print('{0} changed into {1}'.format(
                            df_parameter, his_parameter))
    return df
    
def compare_df_parameter_his_parameter(df_parameter, his_parameters, normalize_by_unicode=True, 
                                       include_simularity=False, sequence_simularity=0.82):
    """
    function to apply unicode normalization and similarity checking for two list of columns
    
    Parameters
    ----------
    df_parameter : str
        string of locations/parameter from a dbf of shp file
    his_parameters : list
        list of strings of locations/parameters of his-file to compare against the df value
    normalize_by_unicode : boolean
        inlcude this option to include NFKD unicode compatibility decomposition.
        see: http://unicode.org/reports/tr15/
    include_simluratiy : boolean
        include this option to include Ratcliff/Obershelp pattern recognition
    sequence_simularity : float
        number between 0.0 and 1.0, function as threshold, where only a simularity above 
        this value is mapped
    
    Returns
    df_parameter : str
        df parameter updated with matching values of his parmater/location if any 
    
    """
    for his_parameter in his_parameters:
        if normalize_by_unicode == True and include_simularity == False:
            if caseless_equal(his_parameter, df_parameter) == True:
                print('Normalized unicode {0} matches {1}'.format(
                    df_parameter, his_parameter))            
                df_parameter = his_parameter


        # option to apply only parameter similarity functionality
        if include_simularity == True and normalize_by_unicode == False:
            smty = similar(his_parameter, df_parameter)
            # print(smty)
            if sequence_simularity <= smty < 1.0:
                print('{0} changed into {1}, since similarity is {2}'.format(
                    df_parameter, his_parameter, round(smty,2)))            
                df_parameter = his_parameter

        # option to apply both unicode normalization and similarity functionality
        if include_simularity == True and normalize_by_unicode == True:
            # first check unicode normalization
            if caseless_equal(his_parameter, df_parameter) == True:
                print('Normalized unicode {0} matches {1}'.format(
                    df_parameter, his_parameter))            
                df_parameter = his_parameter

            # if not succeed try similarity
            else:
                smty = similar(normalize_caseless(his_parameter), normalize_caseless(df_parameter))
                #print(smty)
                if sequence_simularity <= smty < 1.0:
                    print('{0} changed into {1}, since similarity is {2}'.format(
                        df_parameter, his_parameter, round(smty,2)))                
                    df_parameter = his_parameter

    return df_parameter    


import os
import multiprocessing as mp
try:
    from pathlib import Path
except:    
    from pathlib2 import Path
import argparse
import sys
import itertools
import geopandas as gpd
from datetime import datetime, timedelta
import copy
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter, FormatStrFormatter
import numpy as np
import pandas as pd
from scipy import optimize
import fire
import warnings
import configparser
from hkvsobekpy.core.utils import *
try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        from tqdm import tqdm_notebook as tqdm # Jupyter notebook or qtconsole        
    elif shell == 'TerminalInteractiveShell':
        from tqdm import tqdm # Terminal running IPython
    else:
        from tqdm import tqdm # Other type (?)
except NameError:
    from tqdm import tqdm # Probably standard Python interpreter
    
#for debugging within notebook    

try:
    from IPython.core.debugger import set_trace    
except:
    pass

# for python 2/3 compatibility
try:
    xrange
except NameError:
    xrange = range    
try:
    get_ipython().magic('matplotlib inline')    
except:
    pass

class __his_class(object):
    class __errors__(object):
        """
        error class met verschillende foutmeldingen
        """
        @staticmethod
        def fileNotFound(path_file):
            raise IOError("File doesn't exist. Your path '{}' is OK?".format(path_file))
        @staticmethod
        def metadataNotSet():
            raise AttributeError('Metadata is not set. Use hkvsobekpy.ReadMetadata(his_file)')
        @staticmethod
        def metadataError():
            raise AttributeError('Metadata could not be set using function ReadMetadata()')
        @staticmethod
        def administratieError():            
            raise AttributeError('Administratieblok van his-file kan niet worden uitgelezen')
        @staticmethod
        def locationNotFound(loc):            
            raise AttributeError("Could not find location '{}'. Is it an existing location?".format(loc))
        @staticmethod
        def longLocationNotFound(error):
            raise AttributeError('There is an error reading the locations in the hia-file. Full error is:\n{}'.format(error))
        @staticmethod
        def longParameterNotFound(error):
            raise AttributeError('There is an error reading the parameters in the hia-file. Full error is:\n{}'.format(error))
        @staticmethod
        def unexpectedT0Error(timestampInfo):            
            raise AttributeError("T0 could not be read. Timestamp info is '{}'.".format(timestampInfo))            
        @staticmethod
        def parameterNotFound(param):            
            raise AttributeError("Could not find parameter '{}'. Is it an existing parameter?".format(param))
        @staticmethod
        def annualmaxError():
            raise ValueError('Voor de MultiWaardenArray functie kan alleen `year` of `none` gebruikt worden als voor de jaarmax_as parameter')
        @staticmethod
        def gewogenGemiddeldeError():
            raise ValueError('Kan gewogen gemiddelde niet bepalen. Vereiste is minimaal 2 punten aan de linker en rechterzijde van de gekozen T')
        def gebeurtenissenError():
            raise ValueError('Meer gebeurtenissen dan N.')
    
    def __init__(self):
        """
        SobekResultaten class. Class met methodes om een .his bestand 
        uit te lezen.
        WaterstandStatistiek class. Class met methodes om een GumbleFit te bepalen
        """
        # initatie van lege immutable objects
        # zie: https://stackoverflow.com/a/22526544/2459096
        # python2.7: obj = argparse.Namespace()
        # python3.x: obj = types.SimpleNamespace()
        self.hisFile = argparse.Namespace()
        self.hisFile.timestampInfo = argparse.Namespace()
        self.hisFile.parameterInfo = argparse.Namespace()
        self.hisFile.locationInfo = argparse.Namespace()       
        self.hiaFile = argparse.Namespace()
        self.hiaFile.parameterInfo = argparse.Namespace()
        self.hiaFile.locationInfo = argparse.Namespace()       
                
    def _leesHeader(self, f):
        """
        Lees de header van het bestand

        Parameters
        ----------
        f : BufferedReader    
            Leesbaar ruw IO object

        Returns
        -------
        header : array
            De offset van de location in bytes    
        """
        header = []
        for i in range(4):    
            header.append(f.read(40).decode('windows-1252'))  
        return header           
        
    def _infoFromHeader(self,timestampInfo):
        """
        Lees informatie uit de header

        Parameters
        ----------
        timestampInfo : str
            De location in string format

        Returns
        -------
        beginDate : datetime 
            De begindatum T0, waarop de andere datetime gebasseerd zijn
        timeStepInterval : str
            De eenheid waarin de stapgrote van de datetime objecten berekend zijn, 
            dit kan zijn s [seconds], m [minutes], h[hours], d[days]
        timeStepFactor : int
            Factor welke op de stapgrote toegepast moet worden
        """

        if len(timestampInfo) == 40:
            year = timestampInfo[4:8]
            month = timestampInfo[9:11]
            day = timestampInfo[12:14]
            hours = timestampInfo[15:17]
            minutes = timestampInfo[18:20]
            seconds = timestampInfo[21:23]

            beginDate = datetime(int(year), int(month), int(day), int(hours), int(minutes), int(seconds))

            timeStepInterval = timestampInfo[-2:-1]
            timeStepFactor = int(timestampInfo[30:-2])
        else:
            self.__errors__.unexpectedT0Error(timestampInfo)

        return beginDate, timeStepInterval, timeStepFactor    

    def _leesHia(self, f):
        """
        Lees hia bestand
        """
                
        try:
            p = f.resolve()
            self.hiaFile.hiaFileName = str(p)            

            config = configparser.ConfigParser()
            config.read(self.hiaFile.hiaFileName)
            try:
                # parse hia locations to list of dict
                locations = [{'index':int(val)-1, 'long name':config['Long Locations'][val]} for val in config['Long Locations']]  
                self.hiaFile.locationInfo.locations = locations
            except KeyError as e:
                self.__errors__.longLocationNotFound()
            
            try:
                # parse hia parameters to list of dict
                parameters = [{'index':int(val)-1, 'long name':config['Long Parameters'][val]} for val in config['Long Parameters']]
                self.hiaFile.parameterInfo.parameters = parameters
            except KeyError as e:
                self.__errors__.longParametersNotFound()

        except:
            # hia doesn't exist return empty objects
            self.hiaFile.locationInfo.locations = []
            self.hiaFile.parameterInfo.parameters = []
    
    def _leesAdmin(self, f):
        """
        Lees het administratie blok
        """
        self.hisFile.header = self._leesHeader(f)

        # lees info van header
        self.hisFile.timestampInfo.beginDate, self.hisFile.timestampInfo.timeStepInterval, self.hisFile.timestampInfo.timeStepFactor = self._infoFromHeader(self.hisFile.header[3])        

        # lees aantal parameters
        self.hisFile.parameterInfo.numPar = np.fromfile(f,np.int,1)[0]    

        # lees aantal locations
        self.hisFile.locationInfo.numLoc = np.fromfile(f,np.int,1)[0]    

        # lambda function to split string
        split_string = lambda x, n: [x[i:i+n] for i in range(0, len(x), n)]
        
        # lees parameters
        self.hisFile.parameterInfo.parameters = []
        param_string = f.read(self.hisFile.parameterInfo.numPar * 20).decode('windows-1252')
        self.hisFile.parameterInfo.parameters = split_string(param_string, 20)

#         for i in range(self.hisFile.parameterInfo.numPar):
#             self.hisFile.parameterInfo.parameters.append(f.read(20).decode('windows-1252'))    

        # Lees locations
        self.hisFile.locationInfo.id = []
        self.hisFile.locationInfo.locations = []
        
        #self.hisFile.locationInfo.locations = np.fromfile(f, np.dtype("i2, V20" ), 2)#, U20"), 1)
        for i in range(self.hisFile.locationInfo.numLoc):            
            self.hisFile.locationInfo.id.append(np.fromfile(f,np.int,1)[0])
            self.hisFile.locationInfo.locations.append(f.read(20).decode('windows-1252').rstrip())
            

        # lees timestamps
        self.hisFile.headerSize = f.tell()
        self.hisFile.locationInfo.numTime = int((self.hisFile.hisFileSize - self.hisFile.headerSize) /
                      (self.hisFile.locationInfo.numLoc * self.hisFile.parameterInfo.numPar + 1) / 4)

        aantalBytesPerStap = int((self.hisFile.hisFileSize - self.hisFile.headerSize) / self.hisFile.locationInfo.numTime)

        byteNr = self.hisFile.headerSize

        self.hisFile.timestampInfo.moments = []
        
        self.hisFile.timestampInfo.offset = []
        for i in range(self.hisFile.locationInfo.numTime):
            f.seek(byteNr)
            moment = np.fromfile(f,np.int,1)[0]

            if self.hisFile.timestampInfo.timeStepInterval == 's':
                self.hisFile.timestampInfo.moments.append(self.hisFile.timestampInfo.beginDate + timedelta(seconds = (float(moment) * self.hisFile.timestampInfo.timeStepFactor)))
            elif self.hisFile.timestampInfo.timeStepInterval == 'm':
                self.hisFile.timestampInfo.moments.append(self.hisFile.timestampInfo.beginDate + timedelta(minutes = (float(moment) * self.hisFile.timestampInfo.timeStepFactor)))
            elif self.hisFile.timestampInfo.timeStepInterval == 'h':
                self.hisFile.timestampInfo.moments.append(self.hisFile.timestampInfo.beginDate + timedelta(hours= (float(moment) * self.hisFile.timestampInfo.timeStepFactor)))
            elif self.hisFile.timestampInfo.timeStepInterval == 'd':
                self.hisFile.timestampInfo.moments.append(self.hisFile.timestampInfo.beginDate + timedelta(days= (float(moment) * self.hisFile.timestampInfo.timeStepFactor)))
            self.hisFile.timestampInfo.offset.append(f.tell())
            byteNr += aantalBytesPerStap
        self.hisFile.timestampInfo.offset = np.array(self.hisFile.timestampInfo.offset, dtype=np.int64)
        self.hisFile.timestampInfo.N = np.array(self.hisFile.timestampInfo.moments).max().year - np.array(self.hisFile.timestampInfo.moments).min().year + 1
        return True

    def _locOffset(self, loc, param):
        """
        Krijg de offset van de parameter

        Parameters
        ----------
        loc : string    
            De location in kwestie
        param : string
            De parameter in kwestie        

        Returns
        -------
        locOffset : int
            De offset van de location in bytes
        """
        #  Zoek de index van de parameter
        try:
            parFound = self.hisFile.parameterInfo.parameters.index(param)
        except :
            self.__errors__.parameterNotFound(param)            

        # Zoek de index/id van de location
        try:
            # removed dictionary
            # locFound = next((item for item in self.hisFile.locationInfo.locations if item['location'] == loc))
            locFound = self.hisFile.locationInfo.locations.index(loc)
        except :
            self.__errors__.locationNotFound(loc)

        locOffset = locFound * self.hisFile.parameterInfo.numPar * 4 + parFound * 4

        return locOffset   
    
    def _mergeHiaHisLocations(self):
        """
        Merge locations his en hia
        """
        df_his_locs = pd.DataFrame(self.hisFile.locationInfo.locations, columns=['his name'])
        
        df_hia_locs = pd.DataFrame(self.hiaFile.locationInfo.locations)
        # only merge if df_hia_locs contains items
        if not df_hia_locs.empty:
            df_hia_locs.set_index('index', inplace=True)
        
            df_locs = pd.merge(df_his_locs, df_hia_locs, left_index=True, right_index=True, how='outer')
            df_locs.loc[df_locs.isnull().any(axis=1), 'long name'] = df_locs[df_locs.isnull().any(axis=1)]['his name']
            
            self.hisFile.locationInfo.locations = df_locs['long name'].tolist()
        
    def _mergeHiaHisParameters(self):
        """
        Merge parameters his en hia
        """
        df_his_pars = pd.DataFrame(self.hisFile.parameterInfo.parameters, columns=['his name'])
        
        df_hia_pars = pd.DataFrame(self.hiaFile.parameterInfo.parameters)
        # only merge if df_hia_pars contains items
        if not df_hia_pars.empty:        
            df_hia_pars.set_index('index', inplace=True)
                    
            df_pars = pd.merge(df_his_pars, df_hia_pars, left_index=True, right_index=True, how='outer')
            df_pars.loc[df_pars.isnull().any(axis=1), 'long name'] = df_pars[df_pars.isnull().any(axis=1)]['his name']
            
            self.hisFile.parameterInfo.parameters = df_pars['long name'].tolist()
            
    def _process_wrapper(self, chunk, values):
        with open(self.hisFile.hisFileName, "rb") as f:
            for offset in chunk:
                seek = f.seek(offset)
                values.append({'offset_val': offset, 'value':np.fromfile(f,np.float32,1)[0]}) 
                
    def _chunkify(self, array, no_chunks):
        chunks = np.array_split(array, no_chunks)
        chunks = [x for x in chunks if x.size > 0]
        return chunks                
        
    def GetLocations(self):
        """
        Get the available locations from the his-file. 
            
        Returns
        -------
        locations : list
            The locations available within the his-file
        """        
        locations = self.hisFile.locationInfo.locations
        return locations          
        
    def KrijgLokaties(self):
        """
        Krijg de locations beschikbaar in de his-file. 
        
        Parameters
        ----------
        his_file : str 
            pad naar his_file
        
        Returns
        -------
        locations : list
            De locations welke bekend zijn binnen het his-bestand
        """
        warnings.warn("this function will deprecate in the future, use function hkvsobekpy.GetLocations")
        locations = self.hisFile.locationInfo.locations
        return locations
    
    def GetParameters(self):
        """
        Get the available parameters from the his-file. 
            
        Returns
        -------
        parameters : list
            The parameters available within the his-file
        """        
        parameters = self.hisFile.parameterInfo.parameters
        return parameters    
    
    def KrijgParameters(self):#, his_file):
        """
        Krijg de parameters beschikbaar in de his-file. 
        
        Parameters
        ----------
        his_file : str 
            pad naar his_file        
        
        Returns
        -------
        parameters : list
            De parameters welke bekend zijn binnen het his-bestand
        """
        warnings.warn("this function will deprecate in the future, use function hkvsobekpy.GetParameters")
        parameters = self.hisFile.parameterInfo.parameters
        return parameters

    def GetTimestamps(self):
        """
        Get the available timestamps from the his-file. 
            
        Returns
        -------
        timestamps : list of datetime objects
            The timestamps available within the his-file
        """
        timestamps = self.hisFile.timestampInfo.moments
        return timestamps

    def KrijgTijdstappen(self):#, his_file):
        """
        Krijg de timestamps beschikbaar in de his-file. 
        
        Parameters
        ----------
        his_file : str 
            pad naar his_file
            
        Returns
        -------
        timestamps : list
            De timestamps welke bekend zijn binnen het his-bestand
        """
        warnings.warn("this function will deprecate in the future, use function hkvsobekpy.GetTimestamps")
        #self.LeesMetadata(his_file)
        timestamps = self.hisFile.timestampInfo.moments
        return timestamps    
    
    def ReadMetadata(self, his_file, hia_file='auto'):
        """
        Initiate file by reading headers. This function can be followed by the function DataFrame()

        Parameters
        ----------
        his_file : str
            Full path to the file
        hia_file : anyOf(path_to_file, 'auto', 'none')
            Choose from 'auto', 'none' or full path of file.
            Default is 'auto', it will try to use his_file.with_suffix('.hia')
            When 'none' is set, it won't use any .hia-file
            Other values will be treated as full path to the .hia-file
        """
        self.hisFile.metaDataIngelezen = False
        myHisFile = Path(his_file)
            
        try:
            p = myHisFile.resolve()
        except:
            # doesn't exist
            self.__errors__.fileNotFound(myHisFile)

        # exists    
        self.hisFile.hisFileName = str(p)            

        # set length filesize (bytes)
        statInfo = os.stat(his_file)
        self.hisFile.hisFileSize = statInfo.st_size

        # read his file meta data
        with open(self.hisFile.hisFileName, "rb") as f:
            try:
                self._leesAdmin(f)

            except:
                self.__errors__.administratieError()

        # continue with hia file, if not 'none'
        if not hia_file=='none':
            # try to parse a hia file
            # on auto-path
            if hia_file == 'auto':
                myHiaFile = Path(his_file).with_suffix('.hia')
            # on full-path
            else:
                myHiaFile = Path(hia_file)
                try:
                    myHiaFile.resolve()
                except:
                    # doesn't exist
                    self.__errors__.fileNotFound(myHiaFile)
            # file exist, read files 
            # merge with locations and parameters where possible
            self._leesHia(myHiaFile)
            self._mergeHiaHisLocations()
            self._mergeHiaHisParameters()                
            
        self.hisFile.metaDataIngelezen = True
        return self
        

    def LeesMetadata(self,his_file, hia_file='auto'):
        """
        Open het bestand

        Parameters
        ----------
        his_file : str
            Volledige pad van het bestand
        hia_file : anyOf(path_to_file, 'auto', 'none')
            Kies uit 'auto', 'none' of het volledig pad naar het bestand.
            Default is 'auto', het zal dan proberen om his_file.with_suffix('.hia') te gebruiken
            Wanneer 'none' gezet wordt zal er geen link met een .hia bestand gelegd worden.
            Anders is het het volledige pad van het bestand

        Returns
        -------
        Str : 'Metadata Ingelezen'
        
        """
        warnings.warn("this function will deprecate in the future, use function hkvsobekpy.ReadMetadata instead")
        self.hisFile.metaDataIngelezen = False
        myHisFile = Path(his_file)


            
        try:
            p = myHisFile.resolve()
        except:
            # doesn't exist
            self.__errors__.fileNotFound(myHisFile)

        # exists    
        self.hisFile.hisFileName = str(p)            

        # set length filesize (bytes)
        statInfo = os.stat(his_file)
        self.hisFile.hisFileSize = statInfo.st_size

        # read his file meta data
        with open(self.hisFile.hisFileName, "rb") as f:
            try:
                self._leesAdmin(f)

            except:
                self.__errors__.administratieError()

        # continue with hia file, if not 'none'
        if not hia_file=='none':
            # try to parse a hia file
            # on auto-path
            if hia_file == 'auto':
                myHiaFile = Path(his_file).with_suffix('.hia')
            # on full-path
            else:
                myHiaFile = Path(hia_file)
                try:
                    myHiaFile.resolve()
                except:
                    # doesn't exist
                    self.__errors__.fileNotFound(myHiaFile)
            # file exist, read files 
            # merge with locations and parameters where possible
            self._leesHia(myHiaFile)
            self._mergeHiaHisLocations()
            self._mergeHiaHisParameters()                
            
        self.hisFile.metaDataIngelezen = True
        return self
    
    def SelectPeriodeWaardenArray(self, df, startMMdd=(1,1), endMMdd=(12,31), jaarmax_as='none'):
        """
        Selecteer op basis van een DataFrame de gebeurtenissen binnen een bepaalde periode (bijv. een groeiseizoen).
        Waarbij er de mogelijkheid is om de gebeurtenissen binnen deze periode te groeperen.
        Pas op: in de `EnkeleWaardenArray` en `MultiWaardenArray` wordt deze functie intern ook aangeroepen.
        
        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame met datetime als index
        startMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de start datum van de periode. Genoemde datum is inclusief
        endMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de eind datum van de periode. Genoemde datum is inclusief
        jaarmax_as : str
            Mogelijkheden om de DataFrame te groeperen op basis van jaar om het jaarmaxima te bepalen. 
            Het jaarmaxima wordt bepaald nadat de slice van de jaarlijkse periode is toegepast. 
            Keuze bestaat uit:            
            'date' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met de exacte datum van deze gebeurtenis
            'year' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met het jaar van de gebeurtenis
            'none' - retourneert alle gebeurtenissen in elk jaar
            
        Returns
        -------
        df : pandas.DataFrame
            DataFrame met datetime als index
            
        Examples
        --------
        Voor bepaling van het groeiseizoen:
        Bijvoorbeeld het jaarlijks groeseizoen loopt van 15 april tot en met 11 oktober
        startMMdd = (4,15)
        endMMdd = (10,11)
        
        betekent, slice elk jaar in april na de 15e inclusief tot aan oktober voor de 11e inclusief
        
        Voor bepaling van het jaarmaxima:
        groupby = 'date'
                    A
        2012-10-06  1501
        2013-04-22  1534
        2014-04-18  1591

        groupby = 'year'
                    A
        2012        1501
        2013        1534
        2014        1591
        """
        self.startMM, self.startdd = startMMdd
        self.endMM, self.enddd = endMMdd    
        
        # https://stackoverflow.com/a/45996897/2459096
        # maak een month_day dataframe van de MultiYear DataFrame
        month_day = pd.concat([
                        df.index.to_series().dt.month, 
                        df.index.to_series().dt.day
                    ], axis=1).apply(tuple, axis=1)        
        
        # selecteer alleen de gebeurtenissen binnen een jaarlijks terugkerende periode 
        df = df[(month_day >= (self.startMM, self.startdd)) & (month_day <= (self.endMM, self.enddd))]

        if jaarmax_as=='year':
            # groupby year en selecteer de max. 
            # hetvolgende retourneert alleen jaren + max, willen van elk jaar de datum + max
            df = df.groupby(df.index.year).max()

        elif jaarmax_as=='date':
            # groupy year en selecteer de max. Returned volle datums waar de max van dat jaar was 
            key = df.columns.levels[1][0]
            level = df.columns.names[1]
            slice_col = df.columns.levels[0][0]

            # next get the year maxima of the remaining gebeurtenissen
            idx = df.xs(key=key, level=level, axis=1).groupby([df.index.year])[slice_col].transform(max) == df.xs(key=key, level=level, axis=1)[slice_col]
            df = df[idx]            
            
            # if there are multiple events in a single year with same value take the first
            df_unique = pd.DataFrame(df.index, columns=['date'])
            df_unique['index'] = df_unique['date'].apply(lambda x:x.year)
            df_unique = df_unique.groupby('index').first()            
            #slice_unique_values = df_unique['date'].values
            #df = df.loc[slice_unique_values]
            ix2 = pd.np.isin(df.index.values, df_unique.date.values)
            df = df[ix2]

        
        elif jaarmax_as=='none':
            #doe niets
            pass
    
        return df

    def EnkeleWaardenArray(self, location, parameter, startMMdd=(1,1), endMMdd=(12,31), jaarmax_as='none'):
        """
        Lees de waarden van een enkele parameter op een enkele location

        Parameters
        ----------
        location : string    
            De location in kwestie (alleen de eerste 21 karakters, anders cutoff)
        parameter : string
            De parameter in kwestie (alleen de eerste 21 karakters, anders cutoff)
        startMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de start datum van de periode. Genoemde datum is inclusief
        endMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de eind datum van de periode. Genoemde datum is inclusief
        jaarmax_as : str
            Mogelijkheden om de DataFrame te groeperen op basis van jaar om het jaarmaxima te bepalen. 
            Het jaarmaxima wordt bepaald nadat de slice van de jaarlijkse periode is toegepast. 
            Keuze bestaat uit:            
            'date' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met de exacte datum van deze gebeurtenis
            'year' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met het jaar van de gebeurtenis
            'none' - retourneert alle gebeurtenissen in elk jaar            

        Returns
        -------
        df : DataFrame
            DataFrame met een multicolumn van parameter en locations met datetime index en 
            de bijbehorende waarden
            
        Examples
        -----------
        location = '1'
        parameter = 'Waterlevel maximum (mNAP)'
        
        Intern worden alle inputs afgekort op [0:20] en ziet de tool het als:
        location = '1'
        parameter = 'Waterlevel maximum ('

        Voor bepaling van het groeiseizoen:
        Bijvoorbeeld het jaarlijks groeseizoen loopt van 15 april tot en met 11 oktober
        startMMdd = (4,15)
        endMMdd = (10,11)
        
        betekent, slice elk jaar in april na de 15e inclusief tot aan oktober voor de 11e inclusief
        
        Voor bepaling van het jaarmaxima:
        groupby = 'date'
                    A
        2012-10-06  1501
        2013-04-22  1534
        2014-04-18  1591

        groupby = 'year'
                    A
        2012        1501
        2013        1534
        2014        1591
        """

        if not hasattr(self.hisFile, 'metaDataIngelezen'):
            self.__errors__.metadataNotSet()
        
        if self.hisFile.metaDataIngelezen == False:
            self.__errors__.metadataError()
        
        loc = location
        param = parameter
        
        paramLocOffset = self._locOffset(loc, param)
        offsets = self.hisFile.timestampInfo.offset + paramLocOffset
        values= []
        
        with open(self.hisFile.hisFileName, "rb") as f: 

            for offset in offsets:               
                seek = f.seek(offset)
                values.append(np.fromfile(f,np.float32,1)[0])

        # maak dataframe
        df = pd.DataFrame(data=values, index=self.hisFile.timestampInfo.moments, columns=[(loc,param)])
        df.columns = pd.MultiIndex.from_tuples(df.columns, names=['location','parameter'])
        
        df = self.SelectPeriodeWaardenArray(df, startMMdd=startMMdd, endMMdd=endMMdd, jaarmax_as=jaarmax_as)

        return df#.T.squeeze()
    
    def MultiWaardenArray(self, locations, parameters, startMMdd=(1,1), endMMdd=(12,31), 
                          jaarmax_as='none', drop_lege_jaren=True):
        """
        Lees de waarden van meerdere parameters op meerdere locations

        Parameters
        ----------
        locations : list    
            Een list van de locations (type: str) in kwestie (alleen de eerste 21 karakters, anders cutoff) 
        parameters : list
            Een list van de parameters in kwestie (alleen de eerste 21 karakters, anders cutoff) 
        startMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de start datum van de periode. Genoemde datum is inclusief
        endMMdd : tuple
            Tuple in het formaat (M,d). Bepaling van de eind datum van de periode. Genoemde datum is inclusief
        jaarmax_as : str
            Mogelijkheden om de DataFrame te groeperen op basis van jaar om het jaarmaxima te bepalen. 
            Het jaarmaxima wordt bepaald nadat de slice van de jaarlijkse periode is toegepast. 
            Keuze bestaat uit:            
            'year' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met het jaar van de gebeurtenis
            'none' - retourneert alle gebeurtenissen in elk jaar  
        drop_lege_jaren : boolean
            Mogelijkheid om de jaren te verwijderen welke geen gebeurtenissen bevatten voor de gegeven selectie. 
            Wanneer dit als `True` is geselecteerd zal het jaar alleen gedropt worden als deze niet bestaat voor 
            alle locations/parameters in de selectie

        Returns
        -------
        df : DataFrame
            DataFrame met een multicolumn van parameter en locations met datetime index en 
            de bijbehorende waarden
            
        Examples
        -----------
        locations = ['1','6']
        parameters = ['Waterlevel maximum (mNAP)']
        
        Intern worden alle inputs afgekort op [0:20] en ziet de tool het als:
        locations = ['1','6']
        parameters = ['Waterlevel maximum (']
        """
        # error checking
        if jaarmax_as not in ('year', 'none'):
            self.__errors__.annualmaxError()
            
        # define empty dataframe
        clmns = pd.MultiIndex.from_product([parameters,locations], names=['parameters','locations'])
        
        # voor een MultiWaardenArray wordt alleen een jaar meegenomen als index
        # het hangt van een seizoen filter af of er gebeurtenissen voor dat jaar zijn. Bepaal dit eerst
        loc0 = self.hisFile.locationInfo.locations[0]
        par0 = self.hisFile.parameterInfo.parameters[0]
        idx = self.EnkeleWaardenArray(loc0, par0, jaarmax_as=jaarmax_as).index
        
        self.hisFile.timestampInfo.N
        
        df_full = pd.DataFrame(index=idx, columns=clmns)
        df_full.index.name='date'
        for ix in clmns:
            df = self.EnkeleWaardenArray(location=ix[1], parameter=ix[0], 
                                         startMMdd=startMMdd, endMMdd=endMMdd, jaarmax_as=jaarmax_as)
            df_full.loc[:,(ix[0],ix[1])] = df.T.squeeze()
        
        # bepaling of lege jaren gedropt moeten worden
        if drop_lege_jaren:
            df_full.dropna(axis=0, how='all', inplace=True)
        
        return df_full
        
    def read_series(self, his_file, location, parameter, his_folder, normalize_by_unicode=True, 
                include_simularity=True, sequence_simularity=0.82, return_matching_parameter=False):
        """
        Extract timeseries from his-file
        
        Parameters
        ----------
        his_file : str
            name of his-file to query (e.g: 'reachseg', 'calcpnt', 'struc')
        location : str
            location to query in his-file
        parameter : str
            parameter to query in his-file
        his_folder : str
            path to folder containing the his_file
        normalize_by_unicode : boolean (default True)
            inlcude this option to include NFKD unicode compatibility decomposition.
            see: http://unicode.org/reports/tr15/
        include_simluratiy : boolean (default True)
            include this option to include Ratcliff/Obershelp pattern recognition
        sequence_simularity : float (default 0.82)
            number between 0.0 and 1.0, function as threshold, where only a simularity above 
            this value is mapped            
        
        Returns
        -------
        df : pandas.DataFrame
            timeseries containing all available timesteps in his-file
        """
        # get metadata from appropriate his-file
        hisfile = self.LeesMetadata(os.path.join(his_folder, '{0}.his'.format(his_file).upper()))
        his_parameters = hisfile.KrijgParameters()
        his_locations = hisfile.KrijgLokaties()
        
        # normalize unicode and/or check simularity of parameters 
        parameter = compare_df_parameter_his_parameter(parameter.ljust(20), his_parameters, normalize_by_unicode, 
                                               include_simularity, sequence_simularity)
                                               
        # get dataframe for single location/parameter combination
        df = self.EnkeleWaardenArray(
            location=location, 
            parameter=parameter, 
            jaarmax_as='none')
        if return_matching_parameter == True:
            return df, parameter
        else:
            return df
        
    def DataFrame(self):
        """
        Extract all timeseries from his-file as pandas.DataFrame. This is quick.
        
        Returns
        -------
        df : pandas.DataFrame
            containing all available timesteps for all locations and parameters in his-file
        """
        # get metadata from appropriate his-file
        no_moments = len(self.hisFile.timestampInfo.moments)
        no_params = self.hisFile.parameterInfo.numPar
        no_locs = self.hisFile.locationInfo.numLoc

        # the byte range goes like:
        # for t in times: for l in locs: for p in params: print (t, l, p)
        with open(self.hisFile.hisFileName, "rb") as f: 
            start = self.hisFile.timestampInfo.offset[0]
            no_values = (no_moments * no_params * no_locs) + no_moments
            seek = f.seek(start-4)    
            values = np.fromfile(f, np.float32, no_values) # np.float32 == np.dtype('f4')

        # reshape on number of timestamps
        values = values.reshape(no_moments, -1)
        # filter by slice as first column contains 4 non-assigned bytes
        values = np.ravel(values[:,1::])
        # reshape to prepare for dataframe
        values = values.reshape((no_params, no_locs, no_moments), order='F')
        values = values.T.reshape(no_moments, -1)              
        
        # create multi-index for the columns
        lable_locs = self.hisFile.locationInfo.locations
        lable_params = self.hisFile.parameterInfo.parameters
        cols = pd.MultiIndex.from_product([lable_locs, lable_params])

        # parse into dataframe
        df = pd.DataFrame(values, index=self.hisFile.timestampInfo.moments, columns=cols)    
        df = df.swaplevel(i=0, j=1, axis=1)
        df.sort_index(axis=1,inplace=True)
        return df


#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hkvsobekpy', 'hkvsobekpy.core', 'hkvsobekpy.io']

package_data = \
{'': ['*'], 'hkvsobekpy.io': ['.ipynb_checkpoints/*']}

setup(name='hkvsobekpy',
      version='1.4',
      description='package for water-statistics and plausibility checker using his- and bui-file',
      author='HKV',
      author_email='hoek@hkv.nl',
      url='https://www.hkv.nl',
      packages=packages,
      package_data=package_data,
     )



import os
try:
    from pathlib import Path
except:    
    from pathlib2 import Path
import types
import argparse
import sys
import itertools
import geopandas as gpd
from datetime import datetime, timedelta
import copy
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter, FormatStrFormatter
import numpy as np
import pandas as pd
from scipy import optimize
import fire
import warnings
import copy

try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        from tqdm import tqdm_notebook as tqdm # Jupyter notebook or qtconsole        
    elif shell == 'TerminalInteractiveShell':
        from tqdm import tqdm # Terminal running IPython
    else:
        from tqdm import tqdm # Other type (?)
except NameError:
    from tqdm import tqdm # Probably standard Python interpreter
    
#for debugging within notebook    

try:
    from IPython.core.debugger import set_trace    
except:
    pass

# for python 2/3 compatibility
try:
    xrange
except NameError:
    xrange = range    
try:
    get_ipython().magic('matplotlib inline')    
except:
    pass
from hkvsobekpy.io.his import __his_class
his_reader = __his_class()

class __waterlevelstat_class(object):

    def __init__(self):
        """
        SobekResultaten class. Class met methodes om een .his bestand 
        uit te lezen.
        WaterstandStatistiek class. Class met methodes om een GumbleFit te bepalen
        """
        # initatie van lege immutable objects
        # zie: https://stackoverflow.com/a/22526544/2459096
        # python2.7: obj = argparse.Namespace()
        # python3.x: obj = types.SimpleNamespace()
    def _initNamespaces(self):
        try:
            self.stats = types.SimpleNamespace()
            self.stats.plotposities = types.SimpleNamespace()
            self.stats.stap1 = types.SimpleNamespace()
            self.stats.stap2 = types.SimpleNamespace()
            self.stats.stap3 = types.SimpleNamespace()
            self.stats.stap4 = types.SimpleNamespace()     
        except Exception as e:
            print(e)
            self.stats = argparse.Namespace()
            self.stats.plotposities = argparse.Namespace()
            self.stats.stap1 = argparse.Namespace()
            self.stats.stap2 = argparse.Namespace()
            self.stats.stap3 = argparse.Namespace()
            self.stats.stap4 = argparse.Namespace()
        return self

    class __errors__(object):
        """
        error class met verschillende foutmeldingen
        """
        @staticmethod
        def gewogenGemiddeldeError():
            raise ValueError('Kan gewogen gemiddelde niet bepalen. Vereiste is minimaal 2 punten aan de linker en rechterzijde van de gekozen T')
        def gebeurtenissenError():
            raise ValueError('Meer gebeurtenissen dan N. Zorg ervoor dat je tijdreeks slechts 1 gebeurtenis per jaar heeft')
        def schrijfTabelError():
            raise ValueError('Kan deze tabel alleen wegschrijven met T10,25,50,100 voor Gumbel en T10 voor gewogen gemiddelde')

    def _best_fit_slope_and_intercept(self,xs,ys):
        try:
            warnings.filterwarnings("error")
            m = (((xs.mean()*ys.mean()) - (xs*ys).mean()) / ((xs.mean()*xs.mean()) - (xs*xs).mean()))
        except RuntimeWarning:
            #set_trace()
            m = 0#np.nan
        b= ys.mean() - m*xs.mean()
        return m, b

    def _squared_error(self,ys_orig,ys_line):
        return sum((ys_line - ys_orig) * (ys_line - ys_orig))

    def _coefficient_of_determination(self,ys_orig,ys_line):
        y_mean_line = [ys_orig.mean() for y in ys_orig]
        squared_error_regr = self._squared_error(ys_orig, ys_line)
        squared_error_y_mean = self._squared_error(ys_orig, y_mean_line)
        try:
            warnings.filterwarnings("error")
            _r_squared = 1 - (squared_error_regr/squared_error_y_mean)
        except RuntimeWarning:
            _r_squared = 'undefined'
        return _r_squared

    def _r_squared(self,xs,ys):
        """
        Bepaalt de r**2 waarde op basis van twee input arrays

        Parameters
        ----------
        xs : numpy.array
            array met orginele waarden
        ys : numpy.array
            array met afgeleide waarden

        Returns
        -------
        r_squared : float
            de afgeleide r**2 waarde
        """
        m, b = self._best_fit_slope_and_intercept(xs,ys)
        regression_line = [(m*x)+b for x in xs]

        r_squared = self._coefficient_of_determination(ys,regression_line)
        return r_squared        
        
    def _igumbelFunc(self, arr, Td, t):
        """
        Gumbel Functie
        
        Parameters
        ----------
        arr : numpy.array
            array met gesorteerde waterstanden
        Td : 
        
        Returns
        -------
        gumbelFuncFunc : numpy.array
            Gumbel Functie
        """
        a,d = arr
        self.stats.plotposities.gumbelFunc = (float(-1.0)* float(a)*np.log(1.0/t * float(Td))+float(d))
        return self.stats.plotposities.gumbelFunc
    
    def _igumbelFuncFit(self, arr, h, Td, t):
        """
        Gumbel Functie
        
        Parameters
        ----------
        arr : numpy.array
            array met gesorteerde waterstanden
        h : int/float
        
        Td : int/float
        
        T : int/float
        
        Returns
        -------
        gumbelFuncFit : numpy.array
            Fitted Gumbel Functie
        """        
        a,d = arr
        self.stats.plotposities.gumbelFuncFit = h-(float(-1.0)* float(a)*np.log(1.0/t * float(Td))+float(d))        
        return self.stats.plotposities.gumbelFuncFit
    
    def _gewogenGemiddelde4TOI(self, Tarray, WSarray, TOI):
        """
        Functie om het gewogen gemiddelde te krijgen op basis van de vier dichtsbijzijnde waarden rondom de T of interest

        Parameters
        ----------
        Tarray : numpy.array
            Array met de terugkeertijden, 
        WSarray : numpy.array
            Array met de parameterwaarden, bijvoorbeeld waterstanden
        TOI : int
            Waarde met de terugkeertijd of interest, bijvoorbeeld 10 voor T=10

        Returns
        -------

        """
        # get the index of the closest value to T10 on its left side, 
        # T is sorted in descending order so reverse array first to get ascending order
        cl = self._find_closest(Tarray[::-1], TOI)    

        # get the indices of the four closest points
        # [..]  *     *      *      *   |     *     *  [..]
        #           cl-2   cl-1    cl  T10  cl+1  cl+2
        #                  ----    --       ----  ----
        try:
            indices = np.array([cl-1,cl,cl+1,cl+2])

            WSarray_sel  = np.take(WSarray[::-1], indices)
            Tarray_sel   = np.take(Tarray[::-1], indices) 
        except:
            self.__errors__.gewogenGemiddeldeError()

        # inverse distance weighting
        weights = 1 / abs(Tarray_sel - TOI)
        weights /= weights.sum()    

        # weight_cl-1 * ws_cl-1  +  weight_cl * ws_cl  +  weight_cl+1 * ws_cl+1  +  weight_cl+2 * ws_cl+2
        WSarray_TOI = np.dot(weights, WSarray_sel)
        return WSarray_TOI, WSarray_sel, Tarray_sel

    def _find_closest(self, Tarray, TOI):
        # T must be sorted (in ascending order)
        idx = Tarray.searchsorted(TOI)
        idx = np.clip(idx, 1, len(Tarray)-1)
        left = Tarray[idx-1]
        right = Tarray[idx]
        idx -= TOI - left < right - TOI
        return idx    
        
    def _calcFi_T(self, len_arr,N, Ggi=0.44, GgN=0.12):
        """
        Bepaal plotpositie jaarmaximum als kans        

        Parameters
        ----------
        len_arr : int
            aantal gebeurtenissen
        N : int
            aantal jaren waarover de Gumble fit bepaald moet worden
        Ggi : float
            Gringgorten plotposities i (standaard: 0.44)
        GgN : float
            Gringgorten plotposities N (standaard: 0.12)

        Returns
        -------
        T : np.array
            array van terugkeertijden
        Fijaar : np.array
            array van plotposities jaarmaxima
            
        Literature
        ----------
        Voor bepaling Gringorten coefficienten zie bijv: http://glossary.ametsoc.org/wiki/Gringorten_plotting_position
        """
        Fi = []        
        T = []
        for i in range(len_arr):
            i += 1
            f = float(1 - ((i - Ggi) / (N+GgN)))
            Fi.append(f)
        try:
            warnings.filterwarnings("error")
            Fijaar = (np.log(Fi) )* -1
        except RuntimeWarning:
            #set_trace()
            pass
        T = 1 / (Fijaar)
        # set set self
        self.stats.plotposities.Fijaar = Fijaar
        self.stats.plotposities.T = T
        return self.stats.plotposities.T, self.stats.plotposities.Fijaar
    
    def PlotFiguur(self,stat_object,out_folder='none'):
        """
        plot figuur
        
        Parameters
        ----------
        ID : str
            locationnaam
        
        ## STEP1 Gumbel Fit voor T25,T50,T100 voor alle buien
        S1_vAmin : int
            venster array minimum
        S1_vAmax : int
            venster array maximum
        S1_T : ndarray
            Terugkeertijden van jaarmaxima gebeurtenissen genomen over gehele jaar
        S1_ws_srt : ndarray
            Waterstanden welke horen bij de terugkeertijden van jaarmaxima gebeurtenissen genomen over gehele jaar
        S1_GumbelT : ndarray
            Terugkeertijden welke horen bij de Gumbel fit (eg, 25,50,100)
        S1_GumbelWS : ndarray
            Waterstanden welke horen bij de terugkeertijden welke horen bij de Gumbel fit (eg, 25,50,100)
        S1_GumbelWS_line : ndarray
            Waterstanden welke horen bij de terugkeertijden welke horen bij de Gumbel fit vallend binnend het venster
        S1_r_squared : float
            r**2 tussen de Gumbel fit en gebeurtenenissed binnen je venster
        
        ## STEP2 Gumbel Fit voor T10 voor alleen zomerbuien
        S2_T : ndarray
            Terugkeertijden van jaarmaxima gebeurtenissen genomen over het groeiseizoen
        S2_ws_srt : ndarray
            Waterstanden welke horen bij de terugkeertijden van jaarmaxima gebeurtenissen genomen over het groeiseizoen
        
        ## STEP4 Gewogen gemiddelde voor T10 voor alleen de zomerbuien
        S4_Tarray_sel_jaar : ndarray
            Terugkeertijden van jaarmaxima gebeurtenissen genomen over het groeiseizoen welke meegenomen zijn voor bepaling gewogen gemiddeld
        S4_WSarray_sel_jaar : ndarray
            Waterstanden welke horen bij terugkeertijden van jaarmaxima gebeurtenissen welke meegenomen zijn voor bepaling gewogen gemiddeld
        TOI : int
            Terugkeertijd of interest voor bepaling van het gewogen gemiddelde (normaal gesproken 10)
        S4_WSarray_TOI_jaar : float
            Waterstand welke hoort bij de TOI
        
        Returns
        -------
        Maakt figuur en slaat deze in de folder waar het script gerund wordt.
        """

        ID = stat_object.stats.stap1.ID
        S1_vAmin = stat_object.stats.stap1.vAmin
        S1_vAmax = stat_object.stats.stap1.vAmax
        S1_T = stat_object.stats.stap1.T
        S1_ws_srt = stat_object.stats.stap1.ws_srt
        S1_GumbelT = stat_object.stats.stap1.GumbelT
        S1_GumbelWS = stat_object.stats.stap1.GumbelWS
        S1_GumbelWS_line = stat_object.stats.stap1.GumbelWS_line
        S1_r_squared = stat_object.stats.stap1.r_squared
        
        S2_T = stat_object.stats.stap2.T
        S2_ws_srt = stat_object.stats.stap2.ws_srt
        
        S3_Tarray_sel_jaar = stat_object.stats.stap3.Tarray_sel_jaar
        S3_WSarray_sel_jaar = stat_object.stats.stap3.WSarray_sel_jaar
        TOI3 = stat_object.stats.stap3.TOI
        S3_WSarray_TOI_jaar = stat_object.stats.stap3.WSarray_TOI_jaar          
        
        S4_Tarray_sel_jaar = stat_object.stats.stap4.Tarray_sel_jaar
        S4_WSarray_sel_jaar = stat_object.stats.stap4.WSarray_sel_jaar
        TOI = stat_object.stats.stap4.TOI
        S4_WSarray_TOI_jaar = stat_object.stats.stap4.WSarray_TOI_jaar  
        
        
        # Voor het figuur mag de TOI niet geplot worden in de gumbel fit. 
        # Kijk of TOI binnen de GumbelT bestaan en verwijder dit punt als zo.
        if TOI in S1_GumbelT:            
            idx = S1_GumbelT.index(TOI)
            S1_GumbelT = np.delete(S1_GumbelT, idx)
            S1_GumbelWS = np.delete(S1_GumbelWS, idx)       
        
        fig=plt.figure(figsize=(13,6))
        ax = fig.add_subplot(111)
        ax.set_xscale('log')

        # plot jaarmaxima gebeurtenissen over gehele jaar ## OPEN CIRKELS
        ax.plot(S1_T, S1_ws_srt, fillstyle='none', label='waterstanden (jaarmaxima)',
                color='#2E589F', linestyle='', marker='o', markeredgewidth=2, markersize=10, markerfacecoloralt='gray')

        # plot maxima gebeurtenissen over de zomerbuien ## DICHTE CIRKELS
        ax.plot(S2_T, S2_ws_srt, fillstyle='full', label='waterstanden (maxima groeiseizoen)',
               color='cornflowerblue', linestyle='', marker='o', markersize=8, markerfacecoloralt='gray', alpha=0.75)

        # plot gumble fit voor T25, T50, T100 in lijn en punt
        ax.plot(S1_T[S1_vAmin:S1_vAmax],S1_GumbelWS_line, '-', color='#3BBB75', label='Gumbel fit (venster)')
        ax.scatter(S1_GumbelT, S1_GumbelWS,s=100, marker='s', facecolors='#3BBB75', zorder=10,label='waterstand op T25, T50 en T100')

        # plot gewogen gemiddelde voor T10 op basis van geheel jaar in punt
        ax.scatter(TOI3, S3_WSarray_TOI_jaar,s=100, marker='s',facecolors='#FF5AC3', zorder=10,label='waterstand op T10 (stedelijk)') # geheel jaar

        # plot gewogen gemiddelde voor T10 op basis van geheel jaar in punt ## HALF-GEVULDE CIRKELS        
        ax.plot(S3_Tarray_sel_jaar,S3_WSarray_sel_jaar, fillstyle='right', label='gebeurtenissen voor bepaling T'+str(TOI3)+' (stedelijk)',
               color='#865FC5', linestyle='none', marker='o', markersize=8, markerfacecoloralt='white')        
        
        # plot gewogen gemiddelde voor T10 op basis van alleen zomerbuien in punt
        ax.scatter(TOI, S4_WSarray_TOI_jaar,s=100, marker='s',facecolors='#CE0002', zorder=10,label='waterstand op T10 (landelijk)') # zomerbuien               
        
        # plot gewogen gemiddelde voor T10 op basis van zomerbuien in punt ## HALF-GEVULDE CIRKELS        
        ax.plot(S4_Tarray_sel_jaar,S4_WSarray_sel_jaar, fillstyle='right', label='gebeurtenissen voor bepaling T'+str(TOI)+ ' (landelijk)',
               color='#F5AC1B', linestyle='none', marker='o', markersize=8, markerfacecoloralt='white')
                 
        # plot r**2 rechtsbovenin
        if type(S1_r_squared) is str:
            label_r_squared = S1_r_squared
        else:
            label_r_squared = str(np.round(S1_r_squared,2))
            
        ax.text(0.975, 0.975, '$r^2$: '+label_r_squared,
                horizontalalignment='right',
                verticalalignment='top',
                transform=ax.transAxes)
        
        # axes settings
        ax.set_xlim(1,300)
        Ymin = min(S1_ws_srt) - 0.05
        Ymax = max(S1_ws_srt) + 0.25
        ax.set_ylim(Ymin,Ymax)
        ax.set_ylabel('Waterstand (m+NAP)')
        ax.set_xlabel('Terugkeertijd (jaren)')
        ax.xaxis.set_major_formatter(FormatStrFormatter('%.0f'))
        ax.grid(True,which='both', axis='both', linestyle='-', color= '0.75', zorder=0)

        # legend settings
        #legend = plt.legend(loc='upper left',prop={'size':10})
        legend = plt.legend(loc=1, bbox_to_anchor=(1.6, 1.01),prop={'size':10})
        legend.get_frame().set_facecolor('white')
        legend.get_frame().set_alpha(1)
        #legend.get_frame().set_linewidth(1)
        legend.get_frame().set_linewidth(0)
        legend.get_frame().set_edgecolor('#8B8B8B')

        # table settings
        col_labels=['Terugkeertijd','Waterstand (m+NAP)']        
        table_vals = pd.DataFrame([['T'+str(i) for i in S1_GumbelT],S1_GumbelWS.round(decimals=2).tolist()]).T.values.tolist()        
        table_vals = [['T'+str(TOI)+' (landelijk)', str(round(S4_WSarray_TOI_jaar,2))]] + table_vals
        table_vals = [['T'+str(TOI3)+' (stedelijk)', str(round(S3_WSarray_TOI_jaar,2))]] + table_vals

        # the rectangle is where I want to place the table
        the_table = ax.table(cellText=table_vals,
                          colWidths = [0.4]*2,
                          colLabels=col_labels,
                          loc='center right',bbox=[1.01, 0.0, 0.5, 0.5])                

        plt.title('location: '+ID)
        plt.tight_layout()
        if out_folder=='none':
            plt.show()
        else:
            out_folder_png = os.path.join(out_folder,'png')
            out_folder_png_file = os.path.join(out_folder_png, 'waterstandstatistiek_loc_'+ID+'.png')
            plt.savefig(out_folder_png_file, bbox_inches='tight', dpi=90, pad_inches=0.2) 
            plt.clf()
            plt.close(fig)
    
    def _gewogenGemiddelde(self, df_enkel, N=109,TOI=10):
        """
        """
        # lengte van array van waterstanden
        Nws = df_enkel.size        
        T, Fijaar = self._calcFi_T(Nws,N)
        
        # kijk eerst of de input dataframe een multi-column heeft en reduceerd indien ja
        try:
            df_enkel = df_enkel.iloc[:,0]
        except:
            pass

        if len(df_enkel.name) == 2:
            enkel_location = df_enkel.name[0]
            enkel_parameter = df_enkel.name[1]
        else: 
            enkel_location = enkel_parameter = 'niet gedefinieerd'          
        
        df_enkel_sorted = df_enkel.sort_values(ascending=False)        
        
        # bepaal gewogen gemiddelde voor parameter waarde voor terugkeertijd of interest op basis van 4 dichtsbijzijnde waarden
        WSarray_TOI, WSarray_sel, Tarray_sel = self._gewogenGemiddelde4TOI(T, df_enkel_sorted.values, TOI)
        return WSarray_TOI, WSarray_sel, Tarray_sel
    
    def _enkeleGumbelFit(self, df_enkel, N, vensterArray=[0,9], GumbelT=[25,50,100.0], Ggi=0.44, GgN=0.12):
        """
        TOI : int
            Terugkeertijd of interest
            
        N : int
            Pas op: de bepaling voor over hoeveel jaren het aantal plotposities berekend moet worden, 
            moet gedaan worden op basis van het aantal mogelijke jaren in de gebeurtenissen reeks. 
            Bepaal dit op een gebeurtenissenreeks alvorens een periode filter is toegepast. 
        """        

        # lengte van array van waterstanden
        Nws = df_enkel.size  
        if Nws > N:
            #set_trace()
            self.__errors__.gebeurtenissenError()        
        T, Fijaar = self._calcFi_T(Nws, N, Ggi, GgN)        
        

        # kijk eerst of de input dataframe een multi-column heeft en reduceerd indien ja
        try:
            df_enkel = df_enkel.iloc[:,0]
        except:
            pass

        if len(df_enkel.name) == 2:
            enkel_location = df_enkel.name[0]
            enkel_parameter = df_enkel.name[1]
        else: 
            enkel_location = enkel_parameter = 'niet gedefinieerd'   
            
        # get min/max van venster array
        vAmin = vensterArray[0]
        vAmax = vensterArray[1]
        
        df_enkel_sorted = df_enkel.sort_values(ascending=False)                
        
        # bepaal de terugkeertijden waarden en afgeleiden op basis van venster array voor Gumbel fit
        dT = np.log(T[vAmin]) - np.log(T[vAmax])
        dh = df_enkel_sorted.iloc[vAmin] - df_enkel_sorted.iloc[vAmax]
        dTpunt = -np.log(T[vAmax])
        dhpunt = (dh/dT)*dTpunt
        a0 = [dh/dT, df_enkel_sorted.iloc[vAmax]+dhpunt]
        
        # apply de gumbel fit functie
        a_out, cov_x, infodict, msg, flag = optimize.leastsq(self._igumbelFuncFit, a0, args = (df_enkel_sorted.iloc[vAmin:vAmax], 1, T[vAmin:vAmax]), full_output=True)
        # a_out holds the paramters of the fitted line [a and b]
        a = a_out[0]
        b = a_out[1]
        # get the locations on the fit corresponding to the required return periods
        GumbelWS = self._igumbelFunc(a_out, 1, np.array(GumbelT))
        # get the locations on the fit corresponding to the slice on all Ts using the vensterArray
        GumbelWS_lineVensterArray = self._igumbelFunc(a_out, 1, T[vAmin:vAmax])
        #set_trace()
        
        xs = df_enkel_sorted.iloc[vAmin:vAmax]
        ys = GumbelWS_lineVensterArray
        r_squared = self._r_squared(xs,ys)
        
        # schrijf weg naar gumbel dataframe
        T_columns = ['T'+str(int(item)) for item in GumbelT]
        df_gumbel = pd.DataFrame(index=[enkel_location], columns=T_columns+['solution'])
        df_gumbel.index.name='location'

        df_gumbel.loc[enkel_location][T_columns] = GumbelWS
        if flag > 4:
            df_gumbel.loc[enkel_location]['solution'] = 'no solution'
        else:
            df_gumbel.loc[enkel_location]['solution'] = 'optimization succesful'  
            
        return (vAmin,vAmax, df_enkel_sorted.values, T, GumbelT, GumbelWS, enkel_location, a,b, GumbelWS_lineVensterArray,r_squared), df_gumbel

    def AfleidingParameters(self, df_enkel, N, vensterArray, GumbelT, TOI, startMMdd=(1,1), endMMdd=(12,31), 
                            jaarmax_as='date', Ggi=0.44, GgN=0.12):
        """
        Voor het afleiden van de waterstanden welke horen bij de verschillende
        terugkeertijden kan gebruik gemaakt worden van plotposities/Gumbel fit en een 
        gewogen gemiddelde.        
        
        Parameters
        ----------
        N : int
            aantal jaren voor bepaling van de plotposities
        vensterArray : array
            het venster welke gebruikt wordt als filter om de gebeurtenissen mee te 
            nemen voor de bepaling van de Gumbel fit. Het venster is een array van 
            twee waarden, vaak wordt [0,10] gekozen, waar 0 overeenkomt met de meest 
            extreme waarde en 10 de op 10 na meest extreme waarde. In ander woorden, 
            in dit geval is het venster de 10 meest extreme waarden.
        GumbelT : array
            een array met terugkeertijden welke meegenomen moet worden als locations 
            waarover een de waterstand moet bepaald worden aan de hand van de 
            Gumbel fit. Een vaak gebruikte array is [10,25,50,100] wat overeenkomt met 
            de T10, T25, T50 en T100
        TOI : int
            Waarde met de terugkeertijd of interest voor het bepalen van het gewogen 
            gemiddelde. Dit wordt gebruikt voor een terugkeertijd welke tenminste 2 
            gebeurtenissen en 2 gebeurtenissen na zicht heeft. Gewoonlijk kan dit 
            gebruikt worden voor het bepalen van de T10, welke soms nog in de 
            'knik' ligt.
        startMMdd : tuple
            Tuple in het formaat (maand,dag). Bepaling van de start datum van de 
            periode. Genoemde datum is inclusief. Voorbeeld (5,15) staat voor 
            maand 5, dag 15.
        endMMdd : tuple
            Tuple in het formaat (maand,dag). Bepaling van de eind datum van de 
            periode. Genoemde datum is inclusief. Voorbeeld (10,15) staat voor 
            maand 10, dag 15.
        jaarmax_as : str
            Mogelijkheden om de DataFrame te groeperen op basis van jaar om het jaarmaxima te bepalen. 
            Het jaarmaxima wordt bepaald nadat de slice van de jaarlijkse periode is toegepast. 
            Keuze bestaat uit:            
            'date' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met de exacte datum van deze gebeurtenis
            'year' - bepaalt de jaarlijkse maxima en geeft de maximale waarde terug met het jaar van de gebeurtenis
            'none' - retourneert alle gebeurtenissen in elk jaar            
        Ggi : float
            Gringgorten plotposities i (standaard: 0.44)
        GgN : float
            Gringgorten plotposities N (standaard: 0.12)          
            
        Returns:
        --------
        object.stats:
        - STEP 1 Gumbel Fit voor geselecteerde T's [eg. T10,T25,T50,T100] voor alle buien
        - STEP 2 Gumbel Fit voor T10 voor alleen zomerbuien
        - STEP 3 Gewogen gemiddelde voor T10 voor alle buien
        - STEP 4 Gewogen gemiddelde voor T10 voor alleen de zomerbuien
        
        Literature
        ----------
        Voor bepaling Gringorten coefficienten zie bijv: http://glossary.ametsoc.org/wiki/Gringorten_plotting_position            
        """
        self = self._initNamespaces()
        # STEP 1 Gumbel Fit voor geselecteerde T's [eg. T10,T25,T50,T100] voor alle buien
        df_enkel_year_max = his_reader.SelectPeriodeWaardenArray(df_enkel, jaarmax_as='date')
        S1_param4fig, S1_df_gumbel = self._enkeleGumbelFit(df_enkel_year_max, N=N, vensterArray=vensterArray, GumbelT=GumbelT,Ggi=Ggi,GgN=GgN)
        self.stats.stap1.vAmin = S1_param4fig[0]
        self.stats.stap1.vAmax = S1_param4fig[1]
        self.stats.stap1.ws_srt = S1_param4fig[2]
        self.stats.stap1.T = S1_param4fig[3]
        self.stats.stap1.GumbelT = S1_param4fig[4]
        self.stats.stap1.GumbelWS = S1_param4fig[5]
        self.stats.stap1.ID = S1_param4fig[6]
        self.stats.stap1.a = S1_param4fig[7]
        self.stats.stap1.b = S1_param4fig[8]
        self.stats.stap1.GumbelWS_line = S1_param4fig[9]
        self.stats.stap1.r_squared = S1_param4fig[10]
        
        # STEP 2 Gumbel Fit voor T10 voor alleen zomerbuien
        df_enkel_groei = his_reader.SelectPeriodeWaardenArray(df_enkel, startMMdd=startMMdd, endMMdd=endMMdd, jaarmax_as=jaarmax_as)
        S2_param4fig, S2_df_gumbel = self._enkeleGumbelFit(df_enkel_groei, N=N, GumbelT=[TOI], vensterArray = vensterArray,Ggi=Ggi,GgN=GgN)
        self.stats.stap2.ws_srt = S2_param4fig[2]
        self.stats.stap2.T = S2_param4fig[3]
        self.stats.stap2.GumbelT = S2_param4fig[4]
        self.stats.stap2.GumbelWS = S2_param4fig[5]
        self.stats.stap2.startMMdd = startMMdd
        self.stats.stap2.endMMdd = endMMdd
        
        # STEP 3 Gewogen gemiddelde voor T10 voor alle buien
        self.stats.stap3.WSarray_TOI_jaar, self.stats.stap3.WSarray_sel_jaar, self.stats.stap3.Tarray_sel_jaar = self._gewogenGemiddelde(df_enkel_year_max, N=N, TOI=TOI)
        self.stats.stap3.TOI = TOI
        self.stats.stap3.startMMdd = startMMdd
        self.stats.stap3.endMMdd = endMMdd        

        # STEP 4 Gewogen gemiddelde voor T10 voor alleen de zomerbuien
        self.stats.stap4.WSarray_TOI_jaar, self.stats.stap4.WSarray_sel_jaar, self.stats.stap4.Tarray_sel_jaar = self._gewogenGemiddelde(df_enkel_groei, N=N, TOI=TOI)
        self.stats.stap4.TOI = TOI
        self.stats.stap4.startMMdd = startMMdd
        self.stats.stap4.endMMdd = endMMdd
        
        return copy.copy(self)
        
    def EnsembleRunner(self, out_folder, his_file, shp_file, shp_key='nodeID', parameter='auto',
                       startMMdd=(5,15), endMMdd=(10,15), vensterArray=[0,10], GumbelT=[10,25,50,100], 
                       TOI=10, draw_plot=True, write_table=True):
        """
        Lees SOBEK weggeschreven variablene uit en schrijf de afgeleide waterstanden 
        toebehorend aan gebruikers gedefineerde terugkeertijden naar zowel csv/dbf 
        en png. Voor het afleiden van de waterstanden welke horen bij de verschillende
        terugkeertijden kan gebruik gemaakt worden van plotposities/Gumbel fit en een 
        gewogen gemiddelde.

        Input
        -----
        out_folder : path
            een folder waarin de csv en shp/dbf bestand worden weggeschreven.
            voor de figuren moet er ook een folder binnen de out_folder bestaan 
            getiteld 'png'
        his_file : path
            absoluut pad naar een his-file welke ingelezen moet gaan worden
        shp_file : path
            absoluut pad naar een shp-file van waaruit de locations gelinkt moeten 
            worden
        shp_key : str
            string met de naam van de kolom in de shp-file waarin de locations staan
        parameter : str
            naam van de parameter in het his-file welke gebruikt moet gaan worden. 
            Default is 'auto'. In dit geval zal de eerste parameter gebruikt worden 
            vanuit het his-file
        startMMdd : tuple
            Tuple in het formaat (maand,dag). Bepaling van de start datum van de 
            periode. Genoemde datum is inclusief. Voorbeeld (5,15) staat voor 
            maand 5, dag 15.
        endMMdd : tuple
            Tuple in het formaat (maand,dag). Bepaling van de eind datum van de 
            periode. Genoemde datum is inclusief. Voorbeeld (10,15) staat voor 
            maand 10, dag 15.
        vensterArray : array
            het venster welke gebruikt wordt als filter om de gebeurtenissen mee te 
            nemen voor de bepaling van de Gumbel fit. Het venster is een array van 
            twee waarden, vaak wordt [0,10] gekozen, waar 0 overeenkomt met de meest 
            extreme waarde en 10 de op 10 na meest extreme waarde. In ander woorden, 
            in dit geval is het venster de 10 meest extreme waarden.
        GumbelT : array
            een array met terugkeertijden welke meegenomen moet worden als locations 
            waarover een de waterstand moet bepaald worden aan de hand van de 
            Gumbel fit. Een vaak gebruikte array is [10,25,50,100] wat overeenkomt met 
            de T10, T25, T50 en T100
        TOI : int
            Waarde met de terugkeertijd of interest voor het bepalen van het gewogen 
            gemiddelde. Dit wordt gebruikt voor een terugkeertijd welke tenminste 2 
            gebeurtenissen en 2 gebeurtenissen na zicht heeft. Gewoonlijk kan dit 
            gebruikt worden voor het bepalen van de T10, welke soms nog in de 
            'knik' ligt.
        draw_plot : boolean
            Aangeven of je de figuren ook wilt wegschrijven naar png's. Indien zo, 
            moet er een 'png' folder bestaan binnen de out_folder. Voor het figuur 
            zal er gekeken worden of de TOI in de GumbelT bevindt. Indien dit zo is, 
            zal alleen de terugkeertijd van de TOI ingetekend worden.
        write_table : boolean
            Aangeven of je de waarden ook wilt wegschrijven naar tabellen. De tabellen 
            welke worden weggeschreven zijn 1 csv bestand en 1 shp/dbf bestand.

        Return
        ------
        geeft geen return binnen python
        """
        print ('read his-file')
        # create frame table
        his_object = his_reader.LeesMetadata(his_file)
        stat_init = self._initNamespaces()
        stat_table1 = pd.DataFrame()#
        stat_table2 = pd.DataFrame()#self._initTabellen(stat_init,shp_key)
        print ('read shp-file')
        locations_df = gpd.read_file(shp_file)
        
        if parameter=='auto':
            parameter = his_object.hisFile.variabeleInfo.variabelen[0]
        print ('create png-file')
        # iterate locationlist and create item in table
        pbar = tqdm(locations_df[shp_key].tolist())
        for location in pbar:
            pbar.set_description("location %s" % location.ljust(21))
            df_enkel = his_object.EnkeleWaardenArray(location, parameter, startMMdd=(1, 1), endMMdd=(12, 31), jaarmax_as='date')
            if df_enkel.size > his_object.hisFile.tijdstapInfo.N:
                #set_trace()
                self.__errors__.gebeurtenissenError()
            stat_object = self.AfleidingParameters(df_enkel, his_object.hisFile.tijdstapInfo.N, vensterArray, GumbelT, TOI, startMMdd, endMMdd, jaarmax_as='date')
            if draw_plot == True:
                self.PlotFiguur(stat_object, out_folder)
            if write_table ==True:
                stat_tables = self.SchrijfTabellen(stat_object, shp_key=shp_key, init_tabel=False)
                #set_trace()
                stat_table1 = stat_table1.append(stat_tables.stats.df_table1)
                stat_table2 = stat_table2.append(stat_tables.stats.df_table2)
        print ('save csv-file')
        # join on input shapefile and write tables
        # save table 1
        stat_table1.to_csv(os.path.join(out_folder,'waterstandstatistiek_output_1.csv'),index=False)
        print ('save shp-file')
        # prepare table 2 and save
        # set_trace()
        locations_df = locations_df.set_index(shp_key).join(stat_table2.set_index(shp_key)).reset_index()
        locations_df.to_file(os.path.join(out_folder,'waterstandstatistiek_output_2.shp'), driver='ESRI Shapefile')
        stat_table2 = locations_df
        print ('done')

    def _initTabellen(self,stat_object,shp_key):
        """
        Initieer tabelstructuur
        
        Parameters
        ----------
        shp_key : str
            key kolom vanuit de shp/dbf bestand
        """
        stat_object.stats.df_table1 = pd.DataFrame(columns=[shp_key,'H','T','seizoen','methode'])
        stat_object.stats.df_table2 = pd.DataFrame(columns=[shp_key,'T10_LANDELIJK','T10_STEDELIJK',
                                                     'T25','T50','T100','r^2','a','b'])
        return stat_object
        
      
    def SchrijfTabellen(self,stat_object,shp_key,decimals=3, init_tabel=True):
        """
        Schrijf afgeleide parameters naar tabel
        
        Parameters
        ----------
        shp_key : str
            key kolom vanuit de shp/dbf bestand
        decimals : int
            aantal decimalen om af te ronden
        init_tabel : boolean
            parameter om aan te geven of de _initTabellen aangeroepen moet worden. Moet op True staan wanneer functie 
            buiten de SingleRunner functie aangeroepen wordt.
        """
        stat_object._initTabellen(stat_object,shp_key)
        # if init_tabel==True:
            # stat_object._initTabellen(stat_object,shp_key)
        # else:
            # stat_empty_init = self._initNamespaces()
            # stat_empty_table = self._initTabellen(stat_empty_init, shp_key)
            # stat_object.stats.df_table1 = types.SimpleNamespace()
            # stat_object.stats.df_table2 = types.SimpleNamespace()
            # try:
                # stat_object.stats.df_table1 = init_tabel.df_table1
                # stat_object.stats.df_table2 = init_tabel.df_table2
            # except Exception as e:
                # print(e)
                # set_trace()            
        
        table1_list = []
        table2_list = []
        # gewogen gemiddelde voor jaarmaxima alle buien (# STEP 3)
        table1_list.append({shp_key:stat_object.stats.stap1.ID,
                   'H':round(stat_object.stats.stap3.WSarray_TOI_jaar,decimals),
                   'T':stat_object.stats.stap3.TOI,
                   'seizoen':'geheel jaar',
                   'methode':'gewogen gemiddelde'})

        # gewogen gemiddelde voor maxima zomerbuien (# STEP 4)
        table1_list.append({shp_key:stat_object.stats.stap1.ID,
                   'H':round(stat_object.stats.stap4.WSarray_TOI_jaar,decimals),
                   'T':stat_object.stats.stap4.TOI,
                   'seizoen':'alleen zomer',
                   'methode':'gewogen gemiddelde'})

        # Gumbel fit voor enkel maxima zomerbuien (# STEP 2)
        table1_list.append({shp_key:stat_object.stats.stap1.ID,
                   'H':round(stat_object.stats.stap2.GumbelWS[0],decimals),
                   'T':stat_object.stats.stap2.GumbelT[0],
                   'seizoen':'alleen zomer',
                   'methode':'Gumbel fit'})

        # Gumbel fit voor jaarmaxima alle buien (# STEP 1)
        for idx, T in enumerate(stat_object.stats.stap1.GumbelT):
            item = {shp_key:stat_object.stats.stap1.ID,
                    'H':round(stat_object.stats.stap1.GumbelWS[idx],decimals),
                    'T':T,
                    'seizoen':'geheel jaar',
                    'methode':'Gumbel fit'}    
            table1_list.append(item)

        
        # append to dataframe 1
        stat_object.stats.df_table1 = stat_object.stats.df_table1.append(table1_list, ignore_index=True)
        
        if type(stat_object.stats.stap1.r_squared) is not str:
            label_r_squared_table = round(stat_object.stats.stap1.r_squared,decimals)
        else:
            label_r_squared_table = stat_object.stats.stap1.r_squared
        try:
            table2_list.append({shp_key:stat_object.stats.stap1.ID,
                                'T10_LANDELIJK':stat_object.stats.df_table1[(stat_object.stats.df_table1['T'] == 10) & 
                                                (stat_object.stats.df_table1['methode'] == 'gewogen gemiddelde') & 
                                                (stat_object.stats.df_table1['seizoen'] == 'alleen zomer') &
                                                (stat_object.stats.df_table1[shp_key] == stat_object.stats.stap1.ID)]['H'].values[0],
                                'T10_STEDELIJK':stat_object.stats.df_table1[(stat_object.stats.df_table1['T'] == 10) & 
                                                (stat_object.stats.df_table1['methode'] == 'gewogen gemiddelde') & 
                                                (stat_object.stats.df_table1['seizoen'] == 'geheel jaar') &
                                                (stat_object.stats.df_table1[shp_key] == stat_object.stats.stap1.ID)]['H'].values[0],                                            
                                'T25':stat_object.stats.df_table1[(stat_object.stats.df_table1['T'] == 25) & 
                                                (stat_object.stats.df_table1[shp_key] == stat_object.stats.stap1.ID)]['H'].values[0],
                                'T50':stat_object.stats.df_table1[(stat_object.stats.df_table1['T'] == 50) & 
                                                (stat_object.stats.df_table1[shp_key] == stat_object.stats.stap1.ID)]['H'].values[0],
                                'T100':stat_object.stats.df_table1[(stat_object.stats.df_table1['T'] == 100) & 
                                                (stat_object.stats.df_table1[shp_key] == stat_object.stats.stap1.ID)]['H'].values[0],                                
                                'r^2':label_r_squared_table,                                
                                'a':round(stat_object.stats.stap1.a,decimals),
                                'b':round(stat_object.stats.stap1.b,decimals)})    
        except (ValueError,IndexError):
            self.__errors__.schrijfTabelError() 
        # append to dataframe 2
        stat_object.stats.df_table2 = stat_object.stats.df_table2.append(table2_list, ignore_index=True)
        if init_tabel==True:
            return stat_object.stats.df_table1, stat_object.stats.df_table2
        else:        
            return stat_object


import argparse
import pandas as pd
import datetime
import numpy as np
import warnings
# import logging
# from importlib import reload
# # initiatie logging
# reload(logging)
# logging.basicConfig(
    # #filename="{0}/{1}.log".format(logPath, fileName),
    # format='%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s', 
    # level=logging.INFO)
# # to close file:
# #logging.shutdown()


class __bui_class(object):
    class __errors__(object):
        """
        error class met verschillende foutmeldingen
        """
        @staticmethod
        def fileNotFound(file=None):
            raise IOError('Bestand bestaat niet. Is je pad goed? {}'.format(file))        
        @staticmethod
        def variableNotSupported(variable=None):
            raise AttributeError('Variabele {} wordt (nog) niet ondersteund. Zie definition information'.format(variable))
        @staticmethod
        def builengteError(availablevalues=None,expectedvalues=None):
            raise ValueError('Het verwachte aantal tijdstappen {} komt niet overeen met het aanwezige aantal tijdstappen {} in de dataframe'.format(expectedvalues,availablevalues))
#        @staticmethod
#        def metadataNotSet():
#            raise AttributeError('Metadata is niet bekend. Zet met sobek.LeesMetadata(his_file)')
#        @staticmethod
#        def metadataError():
#            raise AttributeError('Metadata kon tijdens de functie LeesMetadata() niet bepaald worden')
#        @staticmethod
#        def administratieError():            
#            raise AttributeError('Administratieblok van his-file kan niet worden uitgelezen')
#        @staticmethod
#        def locationNotFound():            
#            raise AttributeError('location niet gevonden. Is het een bestaande location?')
#        @staticmethod
#        def variabeleNotFound():            
#            raise AttributeError('Parameter niet gevonden. Is het een bestaande parameter?')
#        @staticmethod
#        def jaarmaxError():
#            raise ValueError('Voor de MultiWaardenArray functie kan alleen `year` of `none` gebruikt worden als voor de jaarmax_as parameter')
#        @staticmethod
#        def gewogenGemiddeldeError():
#            raise ValueError('Kan gewogen gemiddelde niet bepalen. Vereiste is minimaal 2 punten aan de linker en rechterzijde van de gekozen T')
#        def gebeurtenissenError():
#            raise ValueError('Meer gebeurtenissen dan N.')
    
    def __init__(self):
        return

    def read_bui(self,filename):
        """
        parse een bui-file naar een dataframe
        Parameters
        ----------
        filename : str
            pad naar het .bui bestand
        
        Returns
        -------
        df : pandas.DataFrame
            dataframe met in de kolommen de neerslag in mm voor de stations
            en in de rijen de waarnemingstijdstappen
        """
        # initatie van lege immutable objects
        # zie: https://stackoverflow.com/a/22526544/2459096
        # python2.7: obj = argparse.Namespace()
        # python3.x: obj = types.SimpleNamespace()        
        self.buiFile = argparse.Namespace()
        with open(filename, 'r') as infile:
            f = infile.readlines()
            for i in range(len(f)):
                #print ('regel {0}'.format(str(i)))
                line = f[i]
                if line[0] == '*':
                    # logging.info('regel {0} is comment'.format(str(i)))
                    
                    if 'GEBRUIK DE DEFAULT DATASET' in line.upper():
                        self.buiFile.default_dataset = int(f[i+1])
                    elif 'AANTAL STATION' in line.upper():
                        self.buiFile.aantal_stations = int(f[i+1])
                    elif 'NAMEN VAN STATION' in line.upper():
                        stations = f[i+1:i+1+self.buiFile.aantal_stations]
                        self.buiFile.stations = [station.rstrip().replace("'","") for station in stations]
                    elif 'AANTAL SECONDEN' in line.upper():
                        geb_sec = f[i+1].split()
                        self.buiFile.aantal_gebeurtenissen = int(geb_sec[0])
                        self.buiFile.aantal_seconden = int(geb_sec[1])
                    elif 'HET FORMAT IS: YYYY' in line.upper():
                        T0_raw = f[i+2]#.split()                
                        idx_block = i+1
            # get data block
            # voor elk station de neerslag in mm per tijdstap
            s = pd.Series(f[idx_block+2:len(f)-1])

        # parse datablock to DataFrame
        df = s.str.rstrip().str.lstrip().str.split(' +', expand=True)
        df = df.apply(pd.to_numeric)            

        # startdatum en -tijd
        T0_str = T0_raw.rstrip().lstrip().split()
        # Het format is: yyyymmdd:hhmmss
        self.buiFile.start_datum        = datetime.datetime(*list(map(int,T0_str[0:5]))) #seconden is eruit gehaald anders str[0:6]
        # lengte van de gebeurtenis in dd hh mm ss
        lengte_gebeurtenis = list(map(int,T0_str[-4::]))
        self.buiFile.delta_dag          = lengte_gebeurtenis[0]
        self.buiFile.delta_uur          = lengte_gebeurtenis[1]
        self.buiFile.delta_minuut       = lengte_gebeurtenis[2]
        self.buiFile.delta_seconde      = lengte_gebeurtenis[3]

        # date_range gebeurtenis
        gebeurtenis_date_range = pd.date_range(
            start=self.buiFile.start_datum, 
            end=None, 
            periods=df.shape[0], 
            freq='{0}S'.format(self.buiFile.aantal_seconden)
        )

        df.set_index(gebeurtenis_date_range, inplace=True)
        df.columns = self.buiFile.stations    

        return df
    
    @staticmethod
    def seconds_toperiods(total_seconds):
        """
        Ontleedt een getal in seconden naar lijst met vaste periodes: days, hours, minutes, seconds
        waarbij day    =          24 * 60 * 60 =     86,400 seconden 
                hour   =               60 * 60 =      3,600 seconden 
                minute =                    60 =         60 seconden
        
        Parameters
        --------------
        total_seconds : float or int 
            bijvoorbeeld Timedelta.total_seconds() van Pandas dataframe bv index[1]-index[0]
        
        Returns
        -------------
        list : list of integer values [days, hours, minutes, seconds]
        """
        
        #years, remainder = divmod(total_seconds,365.25*3600*24) #total seconds per year
        days, remainder = divmod(total_seconds, 24*3600)
        hours, remainder = divmod (remainder, 3600)
        minutes, seconds = divmod(remainder, 60)
        
        return [int(days), int(hours), int(minutes), int(seconds)]
        
    def write_bui(self,df,filename,dataset="bui",comment=None):
        """ 
        Maak van een pandas dataframe een sobek buifile
        
        Parameters
        ----------
        df : pandas dataframe
            index df is datumtijdas. aanname is uniforme tijdstap delta time
            in kolommen neerslag in mm/tijdstap
            kolomnaam = stationnaam
        
        filename: str
            path + naam buifile "{str:8}.bui" lengte buinaam is 8 tekens
        
        dataset: (optional) str [default = "bui"] options ["bui", "reeks"]
            type dataset, dit is voorgedefinieerd door sobek.
            NB. 'reeks' is nog niet ondersteund.
            
        comment: (optional) str
            extra opmerkingenregels die in de header van de buifile worden geplaatst
        
        
        RETURNS
        -----------
        sobekbuifile : gebruiksklare buifile die door sobek2 kan worden ingelezen.
        
        Standaard format sobekbuifile: 
        * user defined comment
        * Gebruik de default dataset voor overige invoer (altijd 1 bij bui, 0 bij reeks)
        1
        * Aantal stations
        2
        * Namen van Stations
        'Station1'
        'Station2'
        * Aantal gebeurtenissen en het aantal seconden per waarnemingsstap
        1 3600
        * Elke commentaarregel wordt begonnen met een * (asteriks).
        * Meteo data: neerslag stations; voor elk station: neerslag intensiteit in mm
        * Eerste record bevat start datum & tijd en lengte van de gebeurtenis in dd hh mm ss
        * Het format is: yyyy mm dd hh mm ss
        * Daarna voor elk station de neerslag in mm per tijdstap.
        2010 11 07 00 00 00 14 00 00 00
        ... space seperated values in [mm/timestep] ... 
        """
        if dataset.upper() != "BUI":
            self.__errors__.variableNotSupported(variabele=dataset)
        
        with open(filename, 'w') as bui:
            
            if not comment==None: 
                #optionele commentaarregel van gebruiker
                bui.write("* "+comment+"\n")
            
            bui.write("* Deze buifile is automatisch aangemaakt door de buifile generator van HKV Lijn in Water"+"\n")
            bui.write("* Gebruik de default dataset voor overige invoer (altijd 1 bij bui, 0 bij reeks)"+"\n")
            if dataset.upper() == "BUI":
                bui.write("1"+"\n")
            if dataset.upper() == "REEKS":
                #throw error
                print("not supported")
            
            bui.write("* Aantal stations"+"\n")
            bui.write(str(df.shape[1])+"\n")
            
            
            bui.write("* Namen van stations"+"\n")
            for station in df.columns:
                bui.write("'{}'".format(str(station))+"\n")
            
            bui.write("* Aantal gebeurtenissen en het aantal seconden per waarnemingsstap"+"\n")
            timedelta = int(datetime.datetime.timestamp(df.index[1]) - datetime.datetime.timestamp(df.index[0]))
            
            
            if dataset.upper() == "BUI": 
                bui.write("1 "+str(timedelta)+"\n")
        
                bui.write("* Elke commentaarregel wordt begonnen met een * (asteriks)."+"\n")
                bui.write("* Meteo data: neerslag stations; voor elk station: neerslag intensiteit in mm."+"\n")
                bui.write("* Eerste record bevat start datum & tijd en lengte van de gebeurtenis in dd hh mm ss"+"\n")
                bui.write("* Het format is: yyyy mm dd hh mm ss"+"\n")
                bui.write("* Daarna voor elk station de neerslag in mm per tijdstap."+"\n")
                
                lengte_bui = self.seconds_toperiods((df.index[-1] - df.index[0]).total_seconds()) 
                bui.write(df.index[0].strftime('%Y %m %d %H %M %S ')+"{:0^2} {:0^2} {:0^2} {:0^2}".format(lengte_bui[0],lengte_bui[1],lengte_bui[2],lengte_bui[3])+"\n")

                
                df.to_csv(bui, sep = " ", index = False, header=False)
                bui.close()

                #controleer tijdas bui
                timesteps =  (df.index[-1] - df.index[0]).total_seconds()/timedelta +1 # +1 om tijdstap t=0 te verreken.
                if int(timesteps) != int(len(df.index)):
                    # ! error ! aantal tijdstappen en builengte komen niet met elkaar overeen
                    self.__errors__.builengteError(availablevalues=int(len(df.index)),expectedvalues=int(timesteps))
                    
          
        
        
        
        
        
        